import os
import shutil

import nglview
import numpy as np
import numpy.testing as npt

from ConservedWaterSearch.utils import (
    read_results,
    visualise_nglview,
    visualise_pymol,
)


def test_visualise_pymol():
    tip_res, resO, resH1, resH2 = read_results(
        "tests/data/Clustering_results.dat",
        "tests/data/Type_Clustering_results.dat",
    )
    tip_res[2] = "WCW"
    tip_res[4] = "HCW"
    visualise_pymol(
        tip_res,
        resO,
        resH1,
        resH2,
        aligned_protein=None,
        output_file="test.pse",
    )
    os.remove("test.pse")


def test_visualise_pymol2():
    visualise_pymol(
        *read_results(
            "tests/data/Clustering_results.dat",
            "tests/data/Type_Clustering_results.dat",
        ),
        aligned_protein=None,
        output_file="test.pdb",
        crystal_waters="3T74",
        ligand_resname="UBY",
    )
    os.remove("test.pdb")


def test_visualise_pymol3():
    visualise_pymol(
        *read_results(
            "tests/data/Clustering_results.dat",
            "tests/data/Type_Clustering_results.dat",
        ),
        aligned_protein=None,
    )


def test_visualise_pymol4():
    visualise_pymol(
        *read_results(
            "tests/data/Clustering_results.dat",
            "tests/data/Type_Clustering_results.dat",
        ),
        aligned_protein="tests/data/aligned.pdb",
    )


def test_visualise_nglview():
    vv = visualise_nglview(
        *read_results(
            "tests/data/Clustering_results.dat",
            "tests/data/Type_Clustering_results.dat",
        ),
        aligned_protein=None,
        crystal_waters="3T74",
    )
    assert type(vv) == nglview.NGLWidget


def test_read_results():
    fle = "tests/data/Clustering_results.dat"
    typefle = "tests/data/Type_Clustering_results.dat"
    water_type, waterO, waterH1, waterH2 = read_results(fle, typefle)
    data = np.loadtxt(fle)
    new_waterO = data[:, :3]
    new_waterH1 = data[:, 3:6]
    new_waterH2 = data[:, 6:9]
    assert water_type == ["FCW"] * 20
    npt.assert_allclose(new_waterO, waterO)
    npt.assert_allclose(new_waterH1, waterH1)
    npt.assert_allclose(new_waterH2, waterH2)


def test_read_results2():
    fle = "tests/data/Clustering_results.dat"
    typefle = "tests/data/Type_Clustering_results.dat"
    newfle = "Clustering_results.dat"
    newtypefle = "Type_Clustering_results.dat"
    water_type, waterO, waterH1, waterH2 = read_results(fle, typefle)
    shutil.copy(fle, newfle)
    shutil.copy(typefle, newtypefle)
    newwater_type, new_waterO, new_waterH1, new_waterH2 = read_results(
        newfle, newtypefle
    )
    os.remove(newfle)
    os.remove(newtypefle)
    assert water_type == newwater_type
    npt.assert_allclose(new_waterO, waterO)
    npt.assert_allclose(new_waterH1, waterH1)
    npt.assert_allclose(new_waterH2, waterH2)



from __future__ import annotations
from typing import TYPE_CHECKING

try:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
except ImportError:
    Axes = Figure = None

try:
    from nglview import NGLWidget
except ImportError:
    NGLWidget = None

if TYPE_CHECKING:
    from io import TextIOWrapper

import os
import warnings
import numpy as np
from sklearn.cluster import OPTICS, cluster_optics_xi, HDBSCAN

from ConservedWaterSearch.hydrogen_orientation import (
    hydrogen_orientation_analysis,
)
from ConservedWaterSearch.utils import (
    read_results,
    visualise_nglview,
    visualise_pymol,
    __check_mpl_installation,
)


class WaterClustering:
    """Class for performing water clustering.

    First, oxygens are clustered using OPTICS or HDBSCAN, followed by
    analysis of orientations for classification of waters into one of 3
    proposed conserved water types (for more information see
    :ref:`conservedwaters:theory, background and methods`):

        - FCW (Fully Conserved Water): hydrogens are strongly oriented in
          two directions with angle of 104.5
        - HCW (Half Conserved Water): one set (cluster) of hydrogens is
          oriented in certain directions and other are spread into different
          orientations with angle of 104.5
        - WCW (Weakly Conserved Water): several orientation combinations
          exsist with satisfying water angles

    To run the calculation use either :py:meth:`multi_stage_reclustering`
    function to start Multi Stage ReClustering (MSRC) procedure or
    :py:meth:`single_clustering` to start a single clustering (SC) procedure.
    MSRC produces better results at the cost of computational time,
    while SC is very quick but results are worse and significant amount of
    waters might not be identified at all. For more details see
    :cite:`conservedwatersearch2022`.

    """

    def __init__(
        self,
        nsnaps: int,
        numbpct_oxygen: float = 0.8,
        normalize_orientations: bool = True,
        save_intermediate_results: bool = True,
        save_results_after_done: bool = True,
        numbpct_hyd_orient_analysis: float = 0.85,
        kmeans_ang_cutoff: float = 120,
        kmeans_inertia_cutoff: float = 0.4,
        FCW_angdiff_cutoff: float = 5,
        FCW_angstd_cutoff: float = 17,
        other_waters_hyd_minsamp_pct: float = 0.15,
        nonFCW_angdiff_cutoff: float = 15,
        HCW_angstd_cutoff: float = 17,
        WCW_angstd_cutoff: float = 20,
        weakly_explained: float = 0.7,
        xiFCW: list[float] = [0.03],
        xiHCW: list[float] = [0.05, 0.01],
        xiWCW: list[float] = [0.05, 0.001],
        njobs: int = 1,
        verbose: int = 0,
        debugO: int = 0,
        debugH: int = 0,
        plotend: bool = False,
        plotreach: bool = False,
    ) -> None:
        """Initialise :py:class:`WaterClustering` class.

        The input parameters determine the options for oxygen clustering and
        hydrogen orienataion analysis if applicable.

        Args:
            nsnaps (int): Number of trajectory snapshots related to
                the data set.
            numbpct_oxygen (float, optional): Percentage of
                ``nsnaps`` required for oxygen cluster to be considered
                valid and water conserved. The check is enforced on
                the lower limit ``nsnaps * numbpct_oxygen`` as well as
                the upper limit ``nsnaps * (2-numbpct_oxygen)``.
                Defaults to 0.8.
            normalize_orientations (bool, optional): If orientations
                should be normalised to unit length or not. Defaults to True.
            save_intermediate_results (bool, optional): If ``True``
                intermediate results are saved throught the calculation
                procedure. Defaults to True.
            save_results_after_done (bool, optional): If ``True`` saves
                the results after the calculation has finished to files
                "Clustering_results.dat" and "Type_Clustering_results.dat".
                Defaults to True.
            numbpct_hyd_orient_analysis (float, optional): Minimum
                allowed size of the hydrogen orientation cluster.
                Defaults to 0.85.
            kmeans_ang_cutoff (float, optional): Maximum value of angle (in
                deg) allowed for FCW in kmeans clustering to be considered
                correct water angle. Defaults to 120.
            kmeans_inertia_cutoff (float, optional): upper limit allowed on
                kmeans inertia (measure of spread of data in a cluster).
                Defaults to 0.4.
            FCW_angdiff_cutoff (float, optional): Maximum value of angle (in
                deg) allowed for FCW in OPTICS/HDBSCAN clustering to be
                considered correct water angle. Defaults to 5.
            FCW_angstd_cutoff (float, optional): Maximal standard deviation
                of angle distribution of orientations of two hydrogens
                allowed for water to be considered FCW. Defaults to 17.
            other_waters_hyd_minsamp_pct (float, optional): Minimum samples to
                choose for OPTICS or HDBSCAN clustering as percentage of
                number of water molecules considered for HCW and WCW.
                Defaults to 0.15.
            nonFCW_angdiff_cutoff (float, optional): Maximum standard
                deviation of angle allowed for HCW and WCW to be considered
                correct water angle. Defaults to 15.
            HCW_angstd_cutoff (float, optional): Maximum standard deviation
                cutoff for WCW angles to be considered correct water angles.
                Defaults to 17.
            WCW_angstd_cutoff (float, optional): Maximum standard deviation
                cutoff for WCW angles to be considered correct water angles.
                Defaults to 20.
            weakly_explained (float, optional): percentage of explained
                hydrogen orientations for water to be considered WCW.
                Defaults to 0.7.
            xiFCW (list, optional): Xi value for OPTICS clustering for
                FCW. Don't touch this unless you know what you are
                doing. Defaults to [0.03].
            xiHCW (list, optional): Xi value for OPTICS clustering for
                HCW. Don't touch this unless you know what you are doing.
                Defaults to [0.05, 0.01].
            xiWCW (list, optional): Xi value for OPTICS clustering for
                WCW. Don't touch this unless you know what you are doing.
                Defaults to [0.05, 0.001].
            njobs (int, optional): how many cpu cores to use for clustering.
                Defaults to 1.
            verbose (int, optional): verbosity of output. Defaults to 0.
            debugH (int, optional): debug level for orientations. Defaults to 0.
            plotreach (bool, optional): weather to plot the reachability
                plot for OPTICS when debuging. Defaults to False.
            which (list, optional): list of strings denoting which water
                types to search for. Allowed is any combination of FCW (fully
                conserved waters), HCW (half conserved waters) and WCW
                (weakly conserved waters). Defaults to ["FCW", "HCW", "WCW"].
        """
        self.numbpct_oxygen = numbpct_oxygen
        self.numbpct_hyd_orient_analysis = numbpct_hyd_orient_analysis
        self.kmeans_ang_cutoff = kmeans_ang_cutoff
        self.kmeans_inertia_cutoff = kmeans_inertia_cutoff
        self.conserved_angdiff_cutoff = FCW_angdiff_cutoff
        self.conserved_angstd_cutoff = FCW_angstd_cutoff
        self.other_waters_hyd_minsamp_pct = other_waters_hyd_minsamp_pct
        self.noncon_angdiff_cutoff = nonFCW_angdiff_cutoff
        self.halfcon_angstd_cutoff = HCW_angstd_cutoff
        self.weakly_angstd_cutoff = WCW_angstd_cutoff
        self.weakly_explained = weakly_explained
        self.xiFCW = xiFCW
        self.xiHCW = xiHCW
        self.xiWCW = xiWCW
        self.njobs = njobs
        self.verbose = verbose
        self.debugO = debugO
        self.debugH = debugH
        self.plotreach = plotreach
        self.normalize_orientations: bool = normalize_orientations
        self.plotend = plotend
        self.save_intermediate_results: bool = save_intermediate_results
        self.save_results_after_done: bool = save_results_after_done
        if self.plotend:
            if not (self.debugH < 2 or self.debugO < 2):
                self.plotend = False
                warnings.warn(
                    "plotend set to True while debugH or debugO are >1; setting back to False"
                )
        self.nsnaps: int = nsnaps
        self._waterO: list[np.ndarray] = []
        self._waterH1: list[np.ndarray] = []
        self._waterH2: list[np.ndarray] = []
        self._water_type: list[str] = []

    def save_clustering_options(
        self,
        clustering_type: str,
        clustering_algorithm: str,
        options: list[int, float],
        whichH: list[str],
        fname: str = "clust_options.dat",
    ) -> None:
        """Function that saves clustering options to a file.

        In order to restart the clustering procedure the intermediate
        results and clustering options need to be known. This function
        enables one to save the clustering options upon the start of
        clustering procedure. This happens automatically depending on
        ``save_intermediate_results`` parameter.

        Args:
            clustering_type (str): type of water clustering procedure to
                employ. The options are "multi_stage_reclustering" and
                "single_clustering".
            clustering_algorithm (str): Clustering algorith to use:
                "OPTICS" or "HDBSCAN"
            options (list): List containing additional options for each
                procedure type. If ``clustering_type`` is
                "single_clustering" options are ``[min_samp, xi]`` where
                xi can be any float if "HDBSCAN" is used. If
                ``clustering_type`` is "multi_stage_reclustering" (or
                the quick version) the options are ``[lower_minsamp_pct,
                every_minsamp, xis]``. Again, in case of "HDBSCAN" xis
                can be a list with any
                single float.
            whichH (list[str]): which water types to seach for. Options
                are "onlyO" or any combination of "FCW", "HCW" and "WCW".
            fname (str, optional): file name to save clustering options
                to. Defaults to "clust_options.dat".
        """
        with open(fname, "w") as f:
            print(self.verbose, file=f)
            print(self.debugO, file=f)
            print(self.debugH, file=f)
            print(self.plotreach, file=f)
            print(self.plotend, file=f)
            print(self.numbpct_oxygen, file=f)
            print(self.numbpct_hyd_orient_analysis, file=f)
            print(self.kmeans_ang_cutoff, file=f)
            print(self.kmeans_inertia_cutoff, file=f)
            print(self.conserved_angdiff_cutoff, file=f)
            print(self.conserved_angstd_cutoff, file=f)
            print(self.other_waters_hyd_minsamp_pct, file=f)
            print(self.noncon_angdiff_cutoff, file=f)
            print(self.halfcon_angstd_cutoff, file=f)
            print(self.weakly_angstd_cutoff, file=f)
            print(self.weakly_explained, file=f)
            print(*self.xiFCW, file=f)
            print(*self.xiHCW, file=f)
            print(*self.xiWCW, file=f)
            print(self.normalize_orientations, file=f)
            print(self.nsnaps, file=f)
            print(clustering_type, file=f)
            print(clustering_algorithm, file=f)
            if type(options) != list:
                raise Exception("option has to be a list")
            for i in options:
                if type(i) != list and type(i) != np.ndarray:
                    print(i, file=f)
                elif type(i) == np.ndarray:
                    print(*list(i), file=f)
                else:
                    print(*i, file=f)
            print(*whichH, file=f)

    def _delete_data(
        self,
        elements: np.ndarray,
        Odata: np.ndarray,
        H1: None | np.ndarray = None,
        H2: None | np.ndarray = None,
    ) -> tuple[np.ndarray, np.ndarray | None, np.ndarray | None]:
        """A helper function for deleting data from the dataset during
        MSRC procedure.

        Args:
            elements (np.ndarray): Indices to delete.
            Odata (np.ndarray): Oxygen data set array of
                Oxygen coordinates which will be cut down.
            H1 (None | np.ndarray, optional): Hydrogen 1
                data set array that contains orientations. Defaults to None.
            H2 (None | np.ndarray, optional): Hydrogen 2
                data set array that contains orientations. Defaults to None.


        Returns:
            tuple[ np.ndarray, np.ndarray | None, np.ndarray | None, ]:
            returns a new set of Oxygen and Hydrogen xyz coordinates array
            with some rows deleted.
        """
        Odata = np.delete(Odata, elements, 0)
        if not (H1 is None):
            H1 = np.delete(H1, elements, 0)
        if not (H2 is None):
            H2 = np.delete(H2, elements, 0)

        if self.save_intermediate_results:
            if not (H1 is None) and not (H2 is None):
                np.savetxt("water_coords_restart1.dat", np.c_[Odata, H1, H2])
            else:
                np.savetxt("water_coords_restart1.dat", np.c_[Odata])
            if os.path.isfile("water_coords_restart1.dat"):
                os.rename("water_coords_restart1.dat", "water_coords_restart.dat")
        return Odata, H1, H2

    def _add_water_solutions(
        self,
        waters: list,
    ) -> None:
        """A helper function which extends the solutions obtained from
        analysing hydrogen orientations.

        Args:
            waters (list): List containing results - coordinates of
                oxygens and two hydrogens and water classification.
        """
        for i in waters:
            self._waterO.append(i[0])
            if len(i) > 2:
                self._waterH1.append(i[1])
                self._waterH2.append(i[2])
            self._water_type.append(i[-1])

    def __check_cls_alg_and_whichH(self, clustering_algorithm, whichH):
        if clustering_algorithm != "OPTICS" and clustering_algorithm != "HDBSCAN":
            raise Exception("clustering algorithm must be OPTICS or HDBSCAN")
        for i in whichH:
            if not (i in ["FCW", "HCW", "WCW", "onlyO"]):
                raise Exception(
                    "whichH supports onlyO or any combination of FCW, HCW and WCW"
                )

    def __check_and_setup_single(self, xis, whichH, clustering_algorithm, minsamp):
        if minsamp is None:
            minsamp = int(self.numbpct_oxygen * self.nsnaps)
        elif type(minsamp) != int:
            raise Exception("minsamp must be an int")
        elif minsamp > self.nsnaps or minsamp <= 0:
            raise Exception("minsamp must be between 0 and nsnaps")
        if xis is None:
            xis = 0.05
        elif type(xis) != float:
            raise Exception("xi must be a float")
        elif xis < 0 or xis > 1:
            raise Exception("xis should be between 0 and 1")
        if self.save_intermediate_results:
            self.save_clustering_options(
                "single_clustering",
                clustering_algorithm,
                [minsamp, xis],
                whichH,
            )
        return minsamp, xis

    def __check_and_setup_MSRC(
        self, lower_minsamp_pct, every_minsamp, xis, whichH, clustering_algorithm
    ):
        for i in xis:
            if type(i) != float:
                raise Exception("xis must contain floats")
            if i > 1 or i < 0:
                raise Exception("xis should be between 0 and 1")
        if lower_minsamp_pct > 1.0000001 or lower_minsamp_pct < 0:
            raise Exception("lower_misamp_pct must be between 0 and 1")
        if type(every_minsamp) != int:
            raise Exception("every_minsamp must be integer")
        if every_minsamp <= 0 or every_minsamp > self.nsnaps:
            raise Exception("every_minsamp must be  0<every_minsamp<=nsnaps")
        minsamps: list = list(
            reversed(
                range(
                    int(self.nsnaps * lower_minsamp_pct),
                    self.nsnaps + 1,
                    every_minsamp,
                )
            )
        )
        if clustering_algorithm == "HDBSCAN":
            lxis: list[float] = [0.0]
            allow_single = False
        elif clustering_algorithm == "OPTICS":
            lxis = xis
            allow_single = None
        if self.save_intermediate_results:
            self.save_clustering_options(
                "multi_stage_reclustering",
                clustering_algorithm,
                [lower_minsamp_pct, every_minsamp, lxis],
                whichH,
            )
        return minsamps, lxis, allow_single

    def __scan_clustering_params(
        self,
        Odata,
        H1,
        H2,
        clustering_algorithm,
        minsamps,
        lxis,
        whichH,
        allow_single,
        restart: bool = True,
    ):
        found: bool = False if len(Odata) < self.nsnaps else True
        for wt in whichH:
            wta = [wt]
            while found:
                found = False
                # loop over minsamps- from N(snapshots) to 0.75*N(snapshots)
                for i in minsamps:
                    if clustering_algorithm == "OPTICS":
                        clust: OPTICS | HDBSCAN = OPTICS(min_samples=int(i), n_jobs=self.njobs)  # type: ignore
                        clust.fit(Odata)
                    # loop over xi
                    for j in lxis:
                        # recalculate reachability - OPTICS reachability has to be recaculated when changing minsamp
                        if clustering_algorithm == "HDBSCAN":
                            clust = HDBSCAN(
                                min_cluster_size=int(self.nsnaps * self.numbpct_oxygen),
                                min_samples=int(i),
                                max_cluster_size=int(
                                    self.nsnaps * (2 - self.numbpct_oxygen)
                                ),
                                cluster_selection_method="eom",
                                n_jobs=self.njobs,
                                allow_single_cluster=allow_single,  # type: ignore
                            )
                            clust.fit(Odata)
                            clusters: np.ndarray = clust.labels_
                        elif clustering_algorithm == "OPTICS":
                            clusters = cluster_optics_xi(
                                reachability=clust.reachability_,  # type: ignore
                                predecessor=clust.predecessor_,  # type: ignore
                                ordering=clust.ordering_,  # type: ignore
                                min_samples=i,
                                xi=j,
                            )[0]
                        # Debug stuff
                        if self.debugO > 0:
                            dbgt: str = ""
                            if self.verbose > 0:
                                (aa, bb) = np.unique(clusters, return_counts=True)
                                dbgt = (
                                    f"Oxygen clustering {type(clust)} minsamp={i}, xi={j}, {len(np.unique(clusters[clusters!=-1]))} clusters \n"
                                    f"Required N(elem) range:{self.nsnaps*self.numbpct_oxygen:.2f} to {(2-self.numbpct_oxygen)*self.nsnaps}; (tar cls size={self.nsnaps} and numbpct={self.numbpct_oxygen:.2f})\n"
                                    f"N(elements) for each cluster: {bb}\n"
                                )
                                print(dbgt)
                            ff: Figure = __oxygen_clustering_plot(
                                Odata, clust, dbgt, self.debugO, self.plotreach
                            )
                        waters, idcs = self._analyze_oxygen_clustering(
                            Odata,
                            H1,
                            H2,
                            clusters,
                            stop_after_frist_water_found=restart,
                            whichH=wta,
                        )
                        if self.debugO == 1:
                            plt = __check_mpl_installation()
                            plt.close(ff)
                        if len(waters) > 0:
                            found = True
                            if clustering_algorithm == "HDBSCAN" and allow_single:
                                allow_single = False
                            Odata, H1, H2 = self._delete_data(idcs, Odata, H1, H2)
                            self._add_water_solutions(waters)
                            if self.save_intermediate_results:
                                self.__save_intermediate_results()
                            i = i - 1
                            break
                    if (found and restart) or len(Odata) < self.nsnaps:
                        break
                # check if size of remaining data set is bigger then number of snapshots
                if (
                    clustering_algorithm == "HDBSCAN"
                    and found is False
                    and allow_single is False
                ):
                    found = True
                    allow_single = True
                if len(Odata) < self.nsnaps:
                    found = False
        if (self.debugH == 1 or self.debugO == 1) and self.plotend:
            plt = __check_mpl_installation()
            plt.show()
        if self.save_results_after_done:
            self.save_results()

    def __save_intermediate_results(self):
        self.save_results(
            fname="Clustering_results_temp1.dat",
            typefname="Type_Clustering_results_temp1.dat",
        )
        if os.path.isfile("Clustering_results_temp1.dat") and os.path.isfile(
            "Type_Clustering_results_temp1.dat"
        ):
            os.rename(
                "Clustering_results_temp1.dat",
                "Clustering_results_temp.dat",
            )
            os.rename(
                "Type_Clustering_results_temp1.dat",
                "Type_Clustering_results_temp.dat",
            )
        else:
            raise Warning(
                "unable to overwrite temp save files. Restarting might not work properly"
            )

    def multi_stage_reclustering(
        self,
        Odata: np.ndarray,
        H1: np.ndarray,
        H2: np.ndarray,
        clustering_algorithm: str = "OPTICS",
        lower_minsamp_pct: float = 0.25,
        every_minsamp: int = 1,
        xis: list[float] = [
            0.1,
            0.05,
            0.01,
            0.005,
            0.001,
            0.0005,
            0.0001,
            0.00001,
        ],
        whichH: list[str] = ["FCW", "HCW", "WCW"],
    ) -> None:
        """Multi Stage ReClustering (MSRC) procedure for obtaining conserved
        water molecules.

        Main loop - loops over water clustering parameter space
        (minsamp and xi) and clusters oxygens first - if a clustering
        with satisfactory oxygen clustering and hydrogen orientation
        clustering (optional) is found, elements of that water cluster
        are removed from the data set and water clustering starts from
        the beginning. Loops until no satisfactory clusterings are
        found. For more details see :cite:`conservedwatersearch2022`.

        Args:
            Odata (np.ndarray): Oxygen coordinates.
            H1 (np.ndarray): Hydrogen 1 orientations.
            H2 (np.ndarray): Hydrogen 2 orientations.
            clustering_algorithm (str, optional): Options are "OPTICS"
                or "HDBSCAN". OPTICS provides slightly better results, but
                is also slightly slower. Defaults to "OPTICS".
            lower_minsamp_pct (float, optional): Lowest minsamp value
                used for clustering. The range is from ``nsnaps``
                to ``lower_minsamp_pct`` times ``nsnaps``.
                Defaults to 0.25.
            every_minsamp (int, optional): Step for sampling of minsamp
                in range from ``nsnaps`` to ``lower_minsamp_pct`` times
                ``nsnaps``. If 1 uses all integer values in range.
                Defaults to 1.
            xis (list[float], optional): List of xis for OPTICS
                clustering. This is ignored for HDBSCAN. Defaults to
                [ 0.1, 0.05, 0.01, 0.005, 0.001, 0.0005, 0.0001, 0.00001].
            whichH (list[str], optional): Defines which water types to
                search for. Any combination of "FCW", "HWC" and "WCW" is
                allowed, or "onlyO" for oxygen clustering only.
                Defaults to ["FCW", "HCW", "WCW"].
        """
        self.__check_cls_alg_and_whichH(clustering_algorithm, whichH)
        minsamps, lxis, allow_single = self.__check_and_setup_MSRC(
            lower_minsamp_pct, every_minsamp, xis, whichH, clustering_algorithm
        )
        self.__scan_clustering_params(
            Odata, H1, H2, clustering_algorithm, minsamps, lxis, whichH, allow_single
        )

    def quick_multi_stage_reclustering(
        self,
        Odata: np.ndarray,
        H1: np.ndarray,
        H2: np.ndarray,
        clustering_algorithm: str = "OPTICS",
        lower_minsamp_pct: float = 0.25,
        every_minsamp: int = 1,
        xis: list[float] = [
            0.1,
            0.05,
            0.01,
            0.005,
            0.001,
            0.0005,
            0.0001,
            0.00001,
        ],
        whichH: list[str] = ["FCW", "HCW", "WCW"],
    ) -> None:
        """Quick Multi Stage ReClustering (QMSRC) procedure for
        obtaining conserved water molecules.

        Main loop - loops over water clustering parameter space
        (minsamp and xi) and clusters oxygens first - clusters with
        satisfactory oxygen clustering and hydrogen orientation
        clustering (optional) are found, elements of those water cluster
        are added to the list of conserved waters. The data for those
        waters is removed from the data set but clustering does not
        restart.

        Args:
            Odata (np.ndarray): Oxygen coordinates.
            H1 (np.ndarray): Hydrogen 1 orientations.
            H2 (np.ndarray): Hydrogen 2 orientations.
            clustering_algorithm (str, optional): Options are "OPTICS"
                or "HDBSCAN". OPTICS provides slightly better results, but
                is also slightly slower. Defaults to "OPTICS".
            lower_minsamp_pct (float, optional): Lowest minsamp value
                used for clustering. The range is from ``nsnaps``
                to ``lower_minsamp_pct`` times ``nsnaps``.
                Defaults to 0.25.
            every_minsamp (int, optional): Step for sampling of minsamp
                in range from ``nsnaps`` to ``lower_minsamp_pct`` times
                ``nsnaps``. If 1 uses all integer values in range.
                Defaults to 1.
            xis (list[float], optional): List of xis for OPTICS
                clustering. This is ignored for HDBSCAN. Defaults to
                [ 0.1, 0.05, 0.01, 0.005, 0.001, 0.0005, 0.0001,
                0.00001].
            whichH (list[str], optional): Defines which water types to
                search for. Any combination of "FCW", "HWC" and "WCW" is
                allowed, or "onlyO" for oxygen clustering only.
                Defaults to ["FCW", "HCW", "WCW"].
        """
        self.__check_cls_alg_and_whichH(clustering_algorithm, whichH)
        minsamps, lxis, allow_single = self.__check_and_setup_MSRC(
            lower_minsamp_pct, every_minsamp, xis, whichH, clustering_algorithm
        )
        self.__scan_clustering_params(
            Odata,
            H1,
            H2,
            clustering_algorithm,
            minsamps,
            lxis,
            whichH,
            allow_single,
            restart=False,
        )

    def single_clustering(
        self,
        Odata: np.ndarray,
        H1: np.ndarray,
        H2: np.ndarray,
        clustering_algorithm: str = "OPTICS",
        minsamp: int | None = None,
        xi: float | None = None,
        whichH: list[str] = ["FCW", "HCW", "WCW"],
    ) -> None:
        """Single clustering procedure.

        In single clustering procedure oxygen clustering is run only
        once with given ``minsamp`` and ``xi`` (if applicable - only for
        OPTICS).

        Args:
            Odata (np.ndarray): Oxygen coordinates.
            H1 (np.ndarray): Hydrogen 1 orientations.
            H2 (np.ndarray): Hydrogen 2 orientations.
            clustering_algorithm (str, optional): Options are "OPTICS"
                or "HDBSCAN". OPTICS provides slightly better results, but
                is also slightly slower. Defaults to "OPTICS".
            minsamp (int | None, optional): Minimum samples parameter
                for OPTICS or HDBSCAN. If None ``numbpct_oxygen`` *
                ``nsnaps`` is used. Defaults to None.
            xi (float | None, optional): Xi value for OPTICS. If
                ``None`` value of 0.05 is used. If
                ``clustering_algorithm`` is HDBSCAN its ignored.
                Defaults to None.
            whichH (list[str], optional): Defines which water types to
                search for. Any combination of "FCW", "HWC" and "WCW" is
                allowed, or "onlyO" for oxygen clustering only.
                Defaults to ["FCW", "HCW", "WCW"].
        """
        self.__check_cls_alg_and_whichH(clustering_algorithm, whichH)
        minsamp, xi = self.__check_and_setup_single(
            xi, whichH, clustering_algorithm, minsamp
        )
        if clustering_algorithm == "OPTICS":
            clust: OPTICS | HDBSCAN = OPTICS(
                min_samples=minsamp,
                min_cluster_size=int(self.numbpct_oxygen * self.nsnaps),
                xi=xi,
                n_jobs=self.njobs,
            )
            clust.fit(Odata)
            clusters = clust.labels_
        if clustering_algorithm == "HDBSCAN":
            clust: OPTICS | HDBSCAN = HDBSCAN(
                min_cluster_size=int(self.nsnaps * self.numbpct_oxygen),
                min_samples=minsamp,
                max_cluster_size=int(self.nsnaps * (2 - self.numbpct_oxygen)),
                cluster_selection_method="eom",
                n_jobs=self.njobs,
            )
            clust.fit(Odata)
            clusters = clust.labels_
        # Debug stuff
        if self.debugO > 0:
            dbgt = ""
            if self.verbose > 0:
                (aa, bb) = np.unique(clusters, return_counts=True)
                dbgt = (
                    f"Oxygen clustering {type(clust)} minsamp={minsamp}, xi={xi}, {len(np.unique(clusters[clusters!=-1]))} clusters \n"
                    f"Required N(elem) range:{self.nsnaps*self.numbpct_oxygen:.2f} to {(2-self.numbpct_oxygen)*self.nsnaps}; (tar cls size={self.nsnaps} and numbpct={self.numbpct_oxygen:.2f})\n"
                    f"N(elements) for each cluster: {bb}\n"
                )
                print(dbgt)
            ff: Figure = __oxygen_clustering_plot(
                Odata, clust, dbgt, self.debugO, self.plotreach
            )
        waters, _ = self._analyze_oxygen_clustering(
            Odata,
            H1,
            H2,
            clusters,
            stop_after_frist_water_found=False,
            whichH=whichH,
        )
        if self.debugO == 1:
            plt = __check_mpl_installation()
            plt.close(ff)
        self._add_water_solutions(waters)
        if self.save_intermediate_results:
            self.__save_intermediate_results()
        if (self.debugH == 1 or self.debugO == 1) and self.plotend:
            plt = __check_mpl_installation()
            plt.show()
        if self.save_results_after_done:
            self.save_results()

    def _analyze_oxygen_clustering(
        self,
        Odata: np.ndarray,
        H1: np.ndarray,
        H2: np.ndarray,
        clusters: np.ndarray,
        stop_after_frist_water_found: bool,
        whichH: list[str],
    ) -> tuple[list[np.ndarray], list[int]]:
        """Helper function for analysing oxygen clustering and invoking
        hydrogen orientation clustering.

        Analyzes clusters for oxygen clustering. For oxygen clusters
        which have the size around number of samples, the hydrogen
        orientation analysis is performed and type of water molecule and
        coordinates are returned.

        Args:
            Odata (np.ndarray): Oxygen coordinates
            H1 (np.ndarray): Hydrogen 1 orientations.
            H2 (np.ndarray): Hydrogen 2 orientations.
            clusters (np.ndarray):  Output of clustering
                results from OPTICS or HDBSCAN.
            stop_after_frist_water_found (bool): If True, the procedure
                is stopped after the first valid water is found and
                indicies of the oxygens and hydrogens belonging to this
                water are also returned.
            whichH (list[str]): Defines which water types to
                search for. Any combination of "FCW", "HWC" and "WCW" is
                allowed, or "onlyO" for oxygen clustering only.

        Returns:
            tuple[list[np.ndarray], list[int]]:
            returns two lists. First list contains valid conserved waters
            found. Each entry in the list is a list which contains the
            positions of oxygen, 2 hydrogen positions and water type
            found, the second list is a list of lists which contain
            arguments to be deleted if ``stop_after_frist_water_found``
            is True, else the second list is empty.
        """
        cluster_ids = np.unique(clusters[clusters != -1])
        cluster_ids.sort()
        min_neioc = self.nsnaps * self.numbpct_oxygen
        max_neioc = self.nsnaps * (2 - self.numbpct_oxygen)
        waters = []
        # make empty numpy array of integers
        idcs = np.array([], dtype=int)
        # Loop over all oxygen clusters (-1 is non cluster)
        for k in cluster_ids:
            mask = (clusters == k)
            # Number of elements in oxygen cluster
            neioc = np.count_nonzero(mask)
            # If number of elements in oxygen cluster is  Nsnap*0.85<Nelem<Nsnap*1.15 then ignore
            if min_neioc < neioc < max_neioc:
                if self.verbose > 0:
                    print(f"O clust {k}, size {len(clusters[clusters==k])}\n")
                O_center = np.mean(Odata[mask], axis=0)
                water = [O_center]
                if not (whichH[0] == "onlyO"):
                    # Construct array of hydrogen orientations
                    orientations = np.vstack([H1[mask], H2[mask]])
                    # Analyse clustering with hydrogen orientation analysis and more debug stuff
                    hyd = hydrogen_orientation_analysis(
                        orientations,
                        self.numbpct_hyd_orient_analysis,
                        self.kmeans_ang_cutoff,
                        self.kmeans_inertia_cutoff,
                        self.conserved_angdiff_cutoff,
                        self.conserved_angstd_cutoff,
                        self.other_waters_hyd_minsamp_pct,
                        self.noncon_angdiff_cutoff,
                        self.halfcon_angstd_cutoff,
                        self.weakly_angstd_cutoff,
                        self.weakly_explained,
                        self.xiFCW,
                        self.xiHCW,
                        self.xiWCW,
                        self.njobs,
                        self.verbose,
                        self.debugH,
                        self.plotreach,
                        whichH,
                        self.normalize_orientations,
                    )
                    if self.plotreach and self.debugH > 0:
                        plt = __check_mpl_installation()
                        plt.show()
                    if len(hyd) > 0:
                        # add water atoms for pymol visualisation
                        for i in hyd:
                            water.append(O_center + i[0])
                            water.append(O_center + i[1])
                            water.append(i[2])
                            waters.append(water)
                        idcs = np.append(idcs, np.argwhere(mask).flatten())
                        # debug
                        if (
                            self.debugO == 1
                            and self.plotreach == 0
                            and self.debugH == 0
                        ):
                            plt = __check_mpl_installation()
                            plt.show()
                        if stop_after_frist_water_found:
                            return waters, idcs
                else:
                    water.append("O_clust")
                    waters.append(water)
                    idcs = np.append(idcs, np.argwhere(mask).flatten())
                    if stop_after_frist_water_found:
                        return waters, idcs
        return waters, idcs

    def restore_default_options(self, delete_results: bool = False) -> None:
        """This function restores all class options to defaults.

        Args:
            delete_results (bool, optional): If ``True`` all the results
                are overwritten with empty lists. Defaults to False.
        """
        self.numbpct_oxygen: float = 0.8
        self.normalize_orientations = True
        self.save_intermediate_results = True
        self.numbpct_hyd_orient_analysis: float = 0.85
        self.kmeans_ang_cutoff: float = 120
        self.kmeans_inertia_cutoff: float = 0.4
        self.conserved_angdiff_cutoff: float = 5
        self.conserved_angstd_cutoff: float = 17
        self.other_waters_hyd_minsamp_pct: float = 0.15
        self.noncon_angdiff_cutoff: float = 15
        self.halfcon_angstd_cutoff: float = 17
        self.weakly_angstd_cutoff: float = 20
        self.weakly_explained: float = 0.7
        self.xiFCW: list = [0.03]
        self.xiHCW: list = [0.05, 0.01]
        self.xiWCW: list = [0.05, 0.001]
        self.njobs: int = 1
        self.verbose: int = 0
        self.debugO: int = 0
        self.debugH: int = 0
        self.plotend: bool = False
        self.plotreach: bool = False
        if delete_results:
            self._water_type = []
            self._waterO = []
            self._waterH1 = []
            self._waterH2 = []

    def save_results(
        self,
        fname: str = "Clustering_results.dat",
        typefname: str = "Type_Clustering_results.dat",
    ) -> None:
        """Saves clustering results to files. Water coordinates are
        saved into the ``fname`` and water types are saved into the
        ``typefname`` file.

        Args:
            fname (str, optional): File name of the file that will
                contain saved water coordinates.
                Defaults to "Clustering_results.dat".
            typefname (str, optional): File name of the file that will
                contain water types.
                Defaults to "Type_Clustering_results.dat".
        """
        if len(self._waterH1) == 0 and len(self._waterH2) == 0:
            np.savetxt(fname, np.c_[self._waterO])
        else:
            np.savetxt(fname, np.c_[self._waterO, self._waterH1, self._waterH2])
        np.savetxt(typefname, np.c_[self._water_type], fmt="%s")

    def restart_cluster(
        self,
        options_file: str = "clust_options.dat",
        data_file: str = "water_coords_restart.dat",
        results_file: str = "Clustering_results_temp.dat",
        type_results_file: str = "Type_Clustering_results_temp.dat",
    ) -> None:
        """Read the clustering options and intermediate results from a
        file and restart the clustering procedure.

        Args:
            options_file (str, optional): File name of the options file.
                Defaults to "clust_options.dat".
            data_file (str, optional): File name of the file containing
                intermediate set of data of hydrogen and oxygen
                coordinates. Defaults to "water_coords_restart.dat".
            results_file (str, optional): File name containing partial
                results with determined water coordinates. Defaults to "Clustering_results_temp.dat".
            type_results_file (str, optional): File name containing
                partial results with water types.
                Defaults to "Type_Clustering_results_temp.dat".
        """
        if os.path.isfile(data_file):
            data: np.ndarray = np.loadtxt(fname=data_file)
            if len(data) == 3:
                Odata: np.ndarray = data
                H1: None | np.ndarray = None
                H2: None | np.ndarray = None
            elif len(data) == 9:
                Odata = data[:, :3]
                H1 = data[:, 3:6]
                H2 = data[:, 6:9]
        else:
            raise Exception("data file not found")
        if os.path.isfile(results_file):
            (
                self._water_type,
                self._waterO,
                self._waterH1,
                self._waterH2,
            ) = read_results(results_file, type_results_file)
        else:
            raise Exception("results file not found")
        self.read_class_options(options_file=options_file)
        (
            clustering_type,
            clustering_algorithm,
            options,
            whichH,
        ) = self.read_water_clust_options(options_file=options_file)
        if clustering_type == "multi_stage_reclustering":
            self.multi_stage_reclustering(
                Odata,  # type: ignore
                H1,  # type: ignore
                H2,  # type: ignore
                clustering_algorithm,
                options[0],
                options[1],  # type: ignore
                options[2],  # type: ignore
                whichH,
            )
        elif clustering_type == "single_clustering":
            self.single_clustering(
                Odata,  # type: ignore
                H1,  # type: ignore
                H2,  # type: ignore
                clustering_algorithm,
                options[0],  # type: ignore
                options[1],
                whichH,
            )
        elif clustering_type == "quick_multi_stage_reclustering":
            self.quick_multi_stage_reclustering(
                Odata,  # type: ignore
                H1,  # type: ignore
                H2,  # type: ignore
                clustering_algorithm,  # type: ignore
                options[0],  # type: ignore
                options[1],  # type: ignore
                options[2],  # type: ignore
                whichH,  # type: ignore
            )
        else:
            raise Exception("incompatible clustering type")

    def read_water_clust_options(
        self, options_file="clust_options.dat"
    ) -> tuple[str, str, list[int | float], list[str]]:
        """Reads clustering procedure type and parameters and returns them.

        Args:
            options_file (str, optional): File to read clustering
                procedure parameters from.
                Defaults to "clust_options.dat".

        Returns:
            Tuple[str, str, list[int | float], list[str]]:
            Returns clustering procedure type, clustering algorithm, options
            for the procedure type and which water types to determine
        """
        if os.path.isfile(options_file):
            f: TextIOWrapper = open(options_file, "r")
            lines: list[str] = f.read().splitlines()
            clustering_type: str = lines[21].strip(" ")
            if (
                clustering_type == "multi_stage_reclustering"
                or clustering_type == "quick_multi_stage_reclustering"
            ):
                clustering_algorithm: str = lines[22].strip(" ")
                lower_minsamp_pct: float = float(lines[23])
                every_minsamp: int = int(lines[24])
                xis: list[float] = [float(i) for i in lines[25].split(" ")]
                whichH: list[str] = [i for i in lines[26].split(" ")]
                options = [lower_minsamp_pct, every_minsamp, xis]
            elif clustering_type == "single_clustering":
                clustering_algorithm = lines[22].strip(" ")
                minsamp: float = float(lines[23])
                xi: float = float(lines[24])
                whichH = [i for i in lines[25].split(" ")]
                options = [minsamp, xi]
            else:
                raise Exception("incompatible clustering type")
            return clustering_type, clustering_algorithm, options, whichH
        else:
            raise Exception("options file not found")

    def read_class_options(self, options_file="clust_options.dat") -> None:
        """Reads all class clustering options from save file and sets
        the parameters. Reads all parameters except clustering protocol
        and protocol parameters.

        Args:
            options_file (str, optional): File name of the options file.
                Defaults to "clust_options.dat".
        """
        if os.path.isfile(options_file):
            f: TextIOWrapper = open(options_file, "r")
            lines: list[str] = f.read().splitlines()
            self.verbose = int(lines[0])
            self.debugO = int(lines[1])
            self.debugH = int(lines[2])
            self.plotreach = lines[3] == "True"
            self.plotend = lines[4] == "True"
            self.numbpct_oxygen = float(lines[5])
            self.numbpct_hyd_orient_analysis = float(lines[6])
            self.kmeans_ang_cutoff = float(lines[7])
            self.kmeans_inertia_cutoff = float(lines[8])
            self.conserved_angdiff_cutoff = float(lines[9])
            self.conserved_angstd_cutoff = float(lines[10])
            self.other_waters_hyd_minsamp_pct = float(lines[11])
            self.noncon_angdiff_cutoff = float(lines[12])
            self.halfcon_angstd_cutoff = float(lines[13])
            self.weakly_angstd_cutoff = float(lines[14])
            self.weakly_explained = float(lines[15])
            self.xiFCW = [float(i) for i in lines[16].split(" ")]
            self.xiHCW = [float(i) for i in lines[17].split(" ")]
            self.xiWCW = [float(i) for i in lines[18].split(" ")]
            self.normalize_orientations = lines[19] == "True"
            self.nsnaps = int(lines[20])
        else:
            raise Exception("options file not found")

    @classmethod
    def create_from_file(
        cls,
        options_file: str = "clust_options.dat",
    ) -> WaterClustering:
        """Create a WaterClustering class from saved clustering options
        file.

        Args:
            options_file (str, optional): File name containig saved
                clustering options. Defaults to "clust_options.dat".

        Returns:
            WaterClustering: creates an instance of :py:class:`WaterClustering`
            class by reading options from a file.
        """
        cls = cls(0)
        if os.path.isfile(options_file):
            f: TextIOWrapper = open(options_file, "r")
            lines: list[str] = f.read().splitlines()
            cls.verbose = int(lines[0])
            cls.debugO = int(lines[1])
            cls.debugH = int(lines[2])
            cls.plotreach = lines[3] == "True"
            cls.plotend = lines[4] == "True"
            cls.numbpct_oxygen = float(lines[5])
            cls.numbpct_hyd_orient_analysis = float(lines[6])
            cls.kmeans_ang_cutoff = float(lines[7])
            cls.kmeans_inertia_cutoff = float(lines[8])
            cls.conserved_angdiff_cutoff = float(lines[9])
            cls.conserved_angstd_cutoff = float(lines[10])
            cls.other_waters_hyd_minsamp_pct = float(lines[11])
            cls.noncon_angdiff_cutoff = float(lines[12])
            cls.halfcon_angstd_cutoff = float(lines[13])
            cls.weakly_angstd_cutoff = float(lines[14])
            cls.weakly_explained = float(lines[15])
            cls.xiFCW = [float(i) for i in lines[16].split(" ")]
            cls.xiHCW = [float(i) for i in lines[17].split(" ")]
            cls.xiWCW = [float(i) for i in lines[18].split(" ")]
            cls.normalize_orientations = lines[19] == "True"
            cls.nsnaps = int(lines[20])
        else:
            raise Exception("options file not found")
        return cls

    def visualise_pymol(
        self,
        aligned_protein: str = "aligned.pdb",
        output_file: str | None = None,
        active_site_ids: list[int] | None = None,
        crystal_waters: str | None = None,
        ligand_resname: str | None = None,
        dist: float = 10.0,
        density_map: str | None = None,
    ) -> None:
        """Visualise results using `pymol <https://pymol.org/>`__.

        Args:
            aligned_protein (str, optional): file name containing protein
                configuration trajectory was aligned to. Defaults to "aligned.pdb".
            output_file (str | None, optional): File to save the
                visualisation state. If ``None``, a pymol session is started
                (this probably doesn't work on Mac OSX). Defaults to None.
            active_site_ids (list[int] | None, optional): Residue ids -
                numbers of aminoacids in active site. These are visualised
                as licorice. Defaults to None.
            crystal_waters (str | None, optional): PDBid from which crystal
                waters will attempted to be extracted. Defaults to None.
            ligand_resname (str | None, optional): Residue name of the
                ligand around which crystal waters (oxygens) shall be
                selected. Defaults to None.
            dist (float): distance from the centre of ligand around which
                crystal waters shall be selected. Defaults to 10.0.
            density_map (str | None, optional): Water density map to add to
                visualisation session (usually .dx file). Defaults to None.
        """
        visualise_pymol(
            self._water_type,
            self._waterO,
            self._waterH1,
            self._waterH2,
            aligned_protein=aligned_protein,
            output_file=output_file,
            active_site_ids=active_site_ids,
            crystal_waters=crystal_waters,
            ligand_resname=ligand_resname,
            dist=dist,
            density_map=density_map,
        )

    def visualise_nglview(
        self,
        aligned_protein: str = "aligned.pdb",
        active_site_ids: list[int] | None = None,
        crystal_waters: str | None = None,
        density_map: str | None = None,
    ) -> NGLWidget:
        """Visualise the results using `nglview <https://github.com/nglviewer/nglview>`__.

        Args:
            aligned_protein (str, optional): File containing protein
                configuration the original trajectory was aligned to.
                Defaults to "aligned.pdb".
            active_site_ids (list[int] | None, optional): Residue ids -
                numbers of aminoacids in active site. These are visualised
                as licorice. Defaults to None.
            crystal_waters (str | None, optional): PDBid from which crystal
                waters will attempted to be extracted. Defaults to None.
            density_map (str | None, optional): Water density map to add to
                visualisation session (usually .dx file). Defaults to None.

        Returns:
            NGLWidget: returns nglview instance widget which can be run
            in Ipyhon/Jupyter to create a visualisation instance
        """
        return visualise_nglview(
            self._water_type,
            self._waterO,
            self._waterH1,
            self._waterH2,
            aligned_protein=aligned_protein,
            active_site_ids=active_site_ids,
            crystal_waters=crystal_waters,
            density_map_file=density_map,
        )

    @property
    def water_type(self) -> list[str]:
        """List containing conserved water type classifications.

        Contains conserved water type classifications in the same order
        as coordinates in ``waterO`` and ``waterH1`` and ``waterH2``.
        Water types:

        - FCW (Fully Conserved Water): hydrogens are strongly oriented in
          two directions with angle of 104.5
        - HCW (Half Conserved Water): one set (cluster) of hydrogens is
          oriented in certain directions and other are spread into different
          orientations with angle of 104.5
        - WCW (Weakly Conserved Water): several orientation combinations
          exsist with satisfying water angles

        For more information see :cite:`conservedwatersearch2022` and
        :ref:`conservedwaters:theory, background and methods`).

        Returns:
            list[str]: Returns a list of strings containing water type
            classification - "FCW" or "HCW" or "WCW". If "onlyO", only
            oxygen clustering was performed.
        """
        return self._water_type

    @property
    def waterO(self) -> list[np.ndarray]:
        """Contains coordiantes of Oxygens of water molecules classified
        with water clustering.

        Returns:
            list[np.ndarray]: Returns a list of 3D xyz
            coordinates of oxygen positions in space
        """
        return self._waterO

    @property
    def waterH1(self) -> list[np.ndarray]:
        """Contains coordinates of first Hydrogen atom of water
        molecules classified with water clustering.

        Returns:
            list[np.ndarray]: Returns a list of 3D xyz
            coordinates of first hydrogens' positions in space
        """
        return self._waterH1

    @property
    def waterH2(self) -> list[np.ndarray]:
        """Contains coordinates of second Hydrogen atom of water
        molecules classified with water clustering.

        Returns:
            list[np.ndarray]: Returns a list of 3D xyz
            coordinates of second hydrogens' positions in space
        """
        return self._waterH2


def __oxygen_clustering_plot(
    Odata: np.ndarray,
    cc: OPTICS,
    title: str,
    debugO: int,
    plotreach: bool = False,
) -> Figure:
    """Function for plotting oxygen clustering results.

    For debuging oxygen clustering. Not ment for general usage.

    """
    if type(cc) != OPTICS:
        plotreach = False
    if debugO > 0:
        plt = __check_mpl_installation()
        fig: Figure = plt.figure()
        if plotreach:
            ax: Axes = fig.add_subplot(1, 2, 1, projection="3d")
        else:
            ax = fig.add_subplot(1, 1, 1, projection="3d")
        for k in np.sort(np.unique(cc.labels_)):
            jaba = Odata[cc.labels_ == k]
            s = 10
            if k == -1:
                s = 0.25
            ax.scatter(
                jaba[:, 0],
                jaba[:, 1],
                jaba[:, 2],
                label=f"{k} ({len(cc.labels_[cc.labels_==k])})",
                s=s,
            )
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.set_title(title)
        ax.legend()
        if plotreach:
            plt = __check_mpl_installation()
            lblls = cc.labels_[cc.ordering_]
            ax = fig.add_subplot(1, 2, 2)
            plt.gca().set_prop_cycle(None)
            space = np.arange(len(Odata))
            ax.plot(space, cc.reachability_)
            for clst in np.unique(lblls):
                if clst == -1:
                    ax.plot(
                        space[lblls == clst],
                        cc.reachability_[lblls == clst],
                        label=f"{clst} ({len(space[lblls==clst])})",
                        color="blue",
                    )
                else:
                    ax.plot(
                        space[lblls == clst],
                        cc.reachability_[lblls == clst],
                        label=f"{clst} ({len(space[lblls==clst])})",
                    )
            ax.legend()
    if debugO == 2:
        plt = __check_mpl_installation()
        plt.show()
    return fig



import os

import numpy as np
import numpy.testing as npt

from ConservedWaterSearch.utils import (
    get_orientations_from_positions,
    read_results,
)
from ConservedWaterSearch.water_clustering import WaterClustering


def test_restoring_defaults() -> None:
    wc = WaterClustering(10)
    wc.restore_default_options()
    wc2 = WaterClustering(10)
    d1 = wc.__dict__
    d2 = wc2.__dict__
    for i in d1.keys():
        assert d1[i] == d2[i]


def test_restoring_defaults2():
    wc = WaterClustering(10)
    wc.debugH = 5
    wc.halfcon_angstd_cutoff = 0.6666
    wc._water_type = ["FCW"]
    wc.restore_default_options(delete_results=True)
    wc2 = WaterClustering(10)
    d1 = wc.__dict__
    d2 = wc2.__dict__
    for i in d1.keys():
        assert d1[i] == d2[i]


def test_save_results1():
    wc = WaterClustering(10)
    wc._waterO.append(np.asarray([0.0, 0.0, 0.0]))
    wc._waterO.append(np.asarray([0.0, 2.0, 0.0]))
    wc._water_type.append("O_clust")
    wc._water_type.append("O_clust")
    wc.save_results()
    a, b, c, d = read_results()
    for i, j in zip(a, wc.water_type):
        assert i == j
    for i, j in zip(b, wc.waterO):
        npt.assert_allclose(i, j)
    os.remove("Clustering_results.dat")
    os.remove("Type_Clustering_results.dat")


def test_save_results2():
    wc = WaterClustering(10)
    wc._waterO.append(np.asarray([0.0, 0.0, 0.0]))
    wc._waterO.append(np.asarray([0.0, 2.0, 0.0]))
    wc._waterH1.append(np.asarray([1.0, 0.0, 0.0]))
    wc._waterH1.append(np.asarray([0.0, 0.8, 0.5]))
    wc._waterH2.append(np.asarray([2.5, 2.0, 2.8]))
    wc._waterH2.append(np.asarray([0.0, 3.0, 0.0]))
    wc._water_type.append("FCW")
    wc._water_type.append("FCW")
    wc.save_results()
    a, b, c, d = read_results()
    for i, j in zip(a, wc.water_type):
        assert i == j
    for i, j in zip(b, wc.waterO):
        npt.assert_allclose(i, j)
    for i, j in zip(c, wc.waterH1):
        npt.assert_allclose(i, j)
    for i, j in zip(d, wc.waterH2):
        npt.assert_allclose(i, j)
    os.remove("Clustering_results.dat")
    os.remove("Type_Clustering_results.dat")


def test_save_clustering_options_and_create_from_file():
    wc = WaterClustering(10)
    ct = "multi_stage_reclustering"
    ca = "OPTICS"
    options = [0.5, 1, [0.05, 0.01, 0.001]]
    whichH = ["onlyO"]
    wc.save_clustering_options(
        clustering_type=ct,
        clustering_algorithm=ca,
        options=options,
        whichH=whichH,
    )
    newWC = WaterClustering.create_from_file()
    d1 = wc.__dict__
    d2 = newWC.__dict__
    for i in d1.keys():
        assert d1[i] == d2[i]
    os.remove("clust_options.dat")


def test_read_clustering_options():
    wc = WaterClustering(10)
    ct = "multi_stage_reclustering"
    ca = "OPTICS"
    options = [0.5, 1, [0.05, 0.01, 0.001]]
    whichH = ["onlyO"]
    wc.save_clustering_options(
        clustering_type=ct,
        clustering_algorithm=ca,
        options=options,
        whichH=whichH,
    )
    newWC = WaterClustering(0)
    newWC.read_class_options()
    d1 = wc.__dict__
    d2 = newWC.__dict__
    for i in d1.keys():
        assert d1[i] == d2[i]
    os.remove("clust_options.dat")


def test_read_water_clust_options():
    wc = WaterClustering(10)
    ct = "multi_stage_reclustering"
    ca = "OPTICS"
    options = [0.5, 1, [0.05, 0.01, 0.001]]
    whichH = ["onlyO"]
    wc.save_clustering_options(
        clustering_type=ct,
        clustering_algorithm=ca,
        options=options,
        whichH=whichH,
    )
    newWC = WaterClustering(0)
    newclstype, newclsalg, newop, newwhich = newWC.read_water_clust_options()
    assert newclstype == ct
    assert newclsalg == ca
    assert newop == options
    assert newwhich == whichH
    os.remove("clust_options.dat")


def test_restart_cluster():
    wc = WaterClustering(10)
    wc._waterO.append(np.asarray([0.0, 0.0, 0.0]))
    wc._waterO.append(np.asarray([0.0, 2.0, 0.0]))
    wc._waterH1.append(np.asarray([1.0, 0.0, 0.0]))
    wc._waterH1.append(np.asarray([0.0, 0.8, 0.5]))
    wc._waterH2.append(np.asarray([2.5, 2.0, 2.8]))
    wc._waterH2.append(np.asarray([0.0, 3.0, 0.0]))
    wc._water_type.append("FCW")
    wc._water_type.append("FCW")
    wc.save_results()
    ct = "multi_stage_reclustering"
    ca = "OPTICS"
    options = [0.5, 1, [0.05, 0.01, 0.001]]
    whichH = ["onlyO"]
    wc.save_clustering_options(
        clustering_type=ct,
        clustering_algorithm=ca,
        options=options,
        whichH=whichH,
    )
    Odata = np.asarray([[0.1, 0.1, 0.1], [1.5, 1.6, 1.7], [1.9, 5.8, 5.6]])
    np.savetxt("water_coords_restart.dat", np.c_[Odata])
    newWC = WaterClustering(0)
    newWC.save_intermediate_results = False
    newWC.restart_cluster(
        results_file="Clustering_results.dat",
        type_results_file="Type_Clustering_results.dat",
    )
    os.remove("clust_options.dat")
    os.remove("Clustering_results.dat")
    os.remove("Type_Clustering_results.dat")


def test_delete_data():
    wc = WaterClustering(0)
    Odata = np.asarray([[0.1, 0.1, 0.1], [1.5, 1.6, 1.7], [1.9, 5.8, 5.6]])
    Onew, _, _ = wc._delete_data([0, 2], Odata=Odata)
    npt.assert_allclose(Onew[0], Odata[1])
    assert os.path.isfile("water_coords_restart.dat")
    os.remove("water_coords_restart.dat")


def test_delete_data2():
    wc = WaterClustering(0)
    Odata = np.asarray([[0.1, 0.1, 0.1], [1.5, 1.6, 1.7], [1.9, 5.8, 5.6]])
    wc.save_intermediate_results = False
    wc._delete_data([0, 2], Odata=Odata)
    assert not (os.path.isfile("water_coords_restart.dat"))


def test_delete_data3():
    wc = WaterClustering(0)
    Odata = np.asarray([[0.1, 0.1, 0.1], [1.5, 1.6, 1.7], [1.9, 5.8, 5.6]])
    H1 = np.asarray([[0.8, 0.5, 0.8], [1.4, 0.6, 3.7], [3.9, 5.1, 5.9]])
    H2 = np.asarray([[0.4, 0.7, 0.1], [1.7, 3.6, 2.7], [1.8, 3.8, 5.1]])
    Onew, H1new, H2new = wc._delete_data([1, 2], Odata, H1, H2)
    npt.assert_allclose(Onew[0], Odata[0])
    npt.assert_allclose(H1new[0], H1[0])
    npt.assert_allclose(H2new[0], H2[0])
    assert os.path.isfile("water_coords_restart.dat")
    os.remove("water_coords_restart.dat")


def test_single_clustering_OPTICS():
    Nsnap = 20
    Opos = np.loadtxt("tests/data/testdataO.dat")
    Hpos = np.loadtxt("tests/data/testdataH.dat")
    wc = WaterClustering(
        Nsnap, save_intermediate_results=False, save_results_after_done=False
    )
    wc.single_clustering(*get_orientations_from_positions(Opos, Hpos))
    a, b, c, d = read_results(
        "tests/data/Single_OPTICS.dat", "tests/data/Single_OPTICS_Type.dat"
    )
    npt.assert_allclose(b, wc.waterO)
    npt.assert_allclose(c, wc.waterH1)
    npt.assert_allclose(d, wc.waterH2)
    assert all(wc.water_type[i] == a[i] for i in range(len(wc.water_type)))


def test_single_clustering_HDBSCAN():
    Nsnap = 20
    Opos = np.loadtxt("tests/data/testdataO.dat")
    Hpos = np.loadtxt("tests/data/testdataH.dat")
    wc = WaterClustering(
        Nsnap, save_intermediate_results=False, save_results_after_done=False
    )
    wc.single_clustering(
        *get_orientations_from_positions(Opos, Hpos), clustering_algorithm="HDBSCAN"
    )
    a, b, c, d = read_results(
        "tests/data/Single_HDBSCAN.dat", "tests/data/Single_HDBSCAN_Type.dat"
    )
    # sort wc.waterO according to b
    # b is xyz coordinates of water Oxygen. Sort them according to x
    b = np.asarray(b)
    b = b[b[:, 0].argsort()]
    c = np.asarray(c)
    c = c[c[:, 0].argsort()]
    d = np.asarray(d)
    d = d[d[:, 0].argsort()]
    # now dothe same for water
    b1 = np.asarray(wc.waterO)
    b1 = b1[b1[:, 0].argsort()]
    c1 = np.asarray(wc.waterH1)
    c1 = c1[c1[:, 0].argsort()]
    d1 = np.asarray(wc.waterH2)
    d1 = d1[d1[:, 0].argsort()]
    npt.assert_allclose(b, b1)
    npt.assert_allclose(c, c1)
    npt.assert_allclose(d, d1)
    assert all(wc.water_type[i] == a[i] for i in range(len(wc.water_type)))


def test_multistage_reclustering_OPTICS():
    Nsnap = 20
    Opos = np.loadtxt("tests/data/testdataO.dat")
    Hpos = np.loadtxt("tests/data/testdataH.dat")
    wc = WaterClustering(
        Nsnap, save_intermediate_results=False, save_results_after_done=False
    )
    wc.multi_stage_reclustering(*get_orientations_from_positions(Opos, Hpos))
    a, b, c, d = read_results(
        "tests/data/MSR_OPTICS.dat", "tests/data/MSR_OPTICS_Type.dat"
    )
    npt.assert_allclose(b, wc.waterO)
    npt.assert_allclose(c, wc.waterH1)
    npt.assert_allclose(d, wc.waterH2)
    assert all(wc.water_type[i] == a[i] for i in range(len(wc.water_type)))


def test_multistage_reclustering_HDBSCAN():
    Nsnap = 20
    Opos = np.loadtxt("tests/data/testdataO.dat")
    Hpos = np.loadtxt("tests/data/testdataH.dat")
    wc = WaterClustering(
        Nsnap, save_intermediate_results=False, save_results_after_done=False
    )
    wc.multi_stage_reclustering(
        *get_orientations_from_positions(Opos, Hpos), clustering_algorithm="HDBSCAN"
    )
    a, b, c, d = read_results(
        "tests/data/MSR_HDBSCAN.dat", "tests/data/MSR_HDBSCAN_Type.dat"
    )
    assert all(wc.water_type[i] == a[i] for i in range(len(wc.water_type)))
    tol = 1e-1
    # b is xyz coordinates of water Oxygen. Sort them according to x
    b = np.asarray(b)
    b = b[b[:, 0].argsort()]
    c = np.asarray(c)
    c = c[c[:, 0].argsort()]
    d = np.asarray(d)
    d = d[d[:, 0].argsort()]
    # now dothe same for water
    b1 = np.asarray(wc.waterO)
    b1 = b1[b1[:, 0].argsort()]
    c1 = np.asarray(wc.waterH1)
    c1 = c1[c1[:, 0].argsort()]
    d1 = np.asarray(wc.waterH2)
    d1 = d1[d1[:, 0].argsort()]
    npt.assert_allclose(b, b1, atol=tol)
    npt.assert_allclose(c, c1, atol=tol)
    npt.assert_allclose(d, d1, atol=tol)


def test_quick_multistage_reclustering_OPTICS():
    Nsnap = 20
    Opos = np.loadtxt("tests/data/testdataO.dat")
    Hpos = np.loadtxt("tests/data/testdataH.dat")
    wc = WaterClustering(
        Nsnap, save_intermediate_results=False, save_results_after_done=False, verbose=2
    )
    wc.quick_multi_stage_reclustering(*get_orientations_from_positions(Opos, Hpos))
    a, b, c, d = read_results(
        "tests/data/MSR_OPTICS.dat", "tests/data/MSR_OPTICS_Type.dat"
    )
    tol = 1e-4
    # b is xyz coordinates of water Oxygen. Sort them according to x
    b = np.asarray(b)
    b = b[b[:, 0].argsort()]
    c = np.asarray(c)
    c = c[c[:, 0].argsort()]
    d = np.asarray(d)
    d = d[d[:, 0].argsort()]
    # now do the same for water
    b1 = np.asarray(wc.waterO)
    b1 = b1[b1[:, 0].argsort()]
    c1 = np.asarray(wc.waterH1)
    c1 = c1[c1[:, 0].argsort()]
    d1 = np.asarray(wc.waterH2)
    d1 = d1[d1[:, 0].argsort()]
    npt.assert_allclose(b, b1, atol=tol)
    npt.assert_allclose(c, c1, atol=tol)
    npt.assert_allclose(d, d1, atol=tol)
    assert all(wc.water_type[i] == a[i] for i in range(len(wc.water_type)))


def test_quick_multistage_reclustering_HDBSCAN():
    Nsnap = 20
    Opos = np.loadtxt("tests/data/testdataO.dat")
    Hpos = np.loadtxt("tests/data/testdataH.dat")
    wc = WaterClustering(
        Nsnap, save_intermediate_results=False, save_results_after_done=False
    )
    wc.quick_multi_stage_reclustering(
        *get_orientations_from_positions(Opos, Hpos), clustering_algorithm="HDBSCAN"
    )
    a, b, c, d = read_results(
        "tests/data/MSR_HDBSCAN.dat", "tests/data/MSR_HDBSCAN_Type.dat"
    )
    assert all(wc.water_type[i] == a[i] for i in range(len(wc.water_type)))
    tol = 1e-1
    # b is xyz coordinates of water Oxygen. Sort them according to x
    b = np.asarray(b)
    b = b[b[:, 0].argsort()]
    c = np.asarray(c)
    c = c[c[:, 0].argsort()]
    d = np.asarray(d)
    d = d[d[:, 0].argsort()]
    # now do the same for water
    b1 = np.asarray(wc.waterO)
    b1 = b1[b1[:, 0].argsort()]
    c1 = np.asarray(wc.waterH1)
    c1 = c1[c1[:, 0].argsort()]
    d1 = np.asarray(wc.waterH2)
    d1 = d1[d1[:, 0].argsort()]
    npt.assert_allclose(b, b1, atol=tol)
    npt.assert_allclose(c, c1, atol=tol)
    npt.assert_allclose(d, d1, atol=tol)



from __future__ import annotations
import os
import platform
import numpy as np
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    try:
        from nglview import NGLWidget
    except ImportError:
        NGLWidget = None


def __check_mpl_installation():
    """Check if matplotlib is installed."""
    try:
        import matplotlib.pyplot as plt
    except ModuleNotFoundError:
        raise Exception(
            "install matplotlib using conda install -c conda-forge matplotlib or pip install matplotlib"
        )
    return plt


def read_results(
    fname: str = "Clustering_results.dat",
    typefname: str = "Type_Clustering_results.dat",
) -> tuple[list[str], list[np.ndarray], list[np.ndarray], list[np.ndarray]]:
    """Read results from files.

    Read results from files and return them in order for further
    processing.

    Args:
        fname (str, optional): File name of the file that contains
            water coordinates. Defaults to "Clustering_results.dat".
        typefname (str, optional): File name of the file that contains
            water classification strings.
            Defaults to "Type_Clustering_results.dat".

    Returns:
        tuple[ list[str], list[np.ndarray], list[np.ndarray], list[np.ndarray] ]:
        returns list of strings which represents water types, and arrays
        of locations of oxygen and two hyrogens. If only oxygens were
        saved returned hydrogen coordinates are empty arrays

    Examples::

        water_types, coord_O, coord_H1, coord_H2 = read_results(
            fname = "Clust_res.dat",
            typefname = "Type_Clust_res.dat",
        )
    """
    water_type = []
    waterO = []
    waterH1 = []
    waterH2 = []
    coords = np.loadtxt(fname)
    if coords.shape[1] == 3:
        for i in coords:
            waterO.append(i)
    else:
        Opos = coords[:, :3]
        H1 = coords[:, 3:6]
        H2 = coords[:, 6:9]
        for i, j, k in zip(Opos, H1, H2):
            waterO.append(i)
            waterH1.append(j)
            waterH2.append(k)
    types = np.loadtxt(typefname, dtype=str)
    for i in types:
        water_type.append(i)
    return (
        water_type,
        list(np.asarray(waterO)),
        list(np.asarray(waterH1)),
        list(np.asarray(waterH2)),
    )


def get_orientations_from_positions(
    coordsO: np.ndarray, coordsH: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Returns orientations from coordinates.

    Calculates relative orientations of hydrogen atoms from their
    positions. The output of this function can be used as input for
    water clustering.

    Args:
        coordsO (np.ndarray): Oxygen coordinates - shape
            (N_waters, 3)
        coordsH (np.ndarray): Hydrogen coordinates - two
            hydrogens bound to the same oxygen have to be placed
            one after another in the array. Shape: (2*N_waters, 3).

    Returns:
        tuple[np.ndarray, np.ndarray, np.ndarray]:
        returns oxygen coordinates array and two hydrogen orientation
        arrays.
    """
    Odata: np.ndarray = np.asarray(coordsO)
    if len(coordsH) > 1:
        H1: np.ndarray = coordsH[::2, :]
        H2: np.ndarray = coordsH[1:, :]
        H2 = H2[::2, :]
        v1: list[np.ndarray] = []
        v2: list[np.ndarray] = []
        for o, h1, h2 in zip(Odata, H1, H2):
            a: np.ndarray = h1 - o
            b: np.ndarray = h2 - o
            if (np.linalg.norm(h1 - o) > 1.2) or (np.linalg.norm(h2 - o) > 1.2):
                raise ValueError(
                    "bad input HO bonds in water longer than 1.2A; value:",
                    np.linalg.norm(h1 - o),
                    np.linalg.norm(h2 - o),
                )
            v1.append(a)
            v2.append(b)
        H1orientdata: np.ndarray = np.asarray(v1)
        H2orientdata: np.ndarray = np.asarray(v2)
        return Odata, H1orientdata, H2orientdata
    else:
        raise Exception("Hydrogen array of wrong length")


def visualise_pymol(
    water_type: list[str],
    waterO: list[list[float]],
    waterH1: list[list[float]],
    waterH2: list[list[float]],
    aligned_protein: str | None = "aligned.pdb",
    output_file: str | None = None,
    active_site_ids: list[int] | None = None,
    crystal_waters: str | None = None,
    ligand_resname: str | None = None,
    dist: float = 10.0,
    density_map: str | None = None,
    lunch_pymol: bool = True,
    reinitialize: bool = True,
) -> None:
    """Visualises results via `pymol <https://pymol.org/>`__.

    Visualises results using pymol in a pymol session or saves to a file.
    On mac OS the interactive pymol session will fail to lunch. If `output_file`
    is `None`, a visalisation state will be saved to
    `pymol_water_visualization.pse` in this case on mac OS.

    Args:
        water_type (list): List containing water type results from
            water clustering.
        waterO (list): Coordiantes of Oxygen atom in water molecules.
        waterH1 (list): Coordinates of Hydrogen1 atom in water molecules.
        waterH2 (list): Coordinates of Hydrogen2 atom in water molecules.
        aligned_protein (str | None, optional): file name containing protein
            configuration trajectory was aligned to. If `None` no
            protein will be shown. Defaults to "aligned.pdb".
        output_file (str | None, optional): File to save the
            visualisation state. If ``None``, a pymol session is started
            (this probably doesn't work on Mac OSX). Defaults to None.
        active_site_ids (list[int] | None, optional): Residue ids -
            numbers of aminoacids in active site. These are visualised
            as licorice. Defaults to None.
        crystal_waters (str | None, optional): PDBid from which crystal
            waters will attempted to be extracted. Defaults to None.
        ligand_resname (str | None, optional): Residue name of the
            ligand around which crystal waters (oxygens) shall be
            selected. Defaults to None.
        dist (float): distance from the centre of ligand around which
            crystal waters shall be selected. Defaults to 10.0.
        density_map (str | None, optional): Water density map to add to
            visualisation session (usually .dx file). Defaults to None.
        lunch_pymol (bool, optional): If `True` pymol will be lunched
            in interactive mode. If `False` pymol will be imported
            without lunching. Defaults to True.
        reinitialize (bool, optional): If `True` pymol will be
            reinitialized (defaults restored and objects cleaned).
            Defaults to True.

    Example::

        # Read the results from files
        water_types, coord_O, coord_H1, coord_H2 = read_results(
            fname = "Clust_res.dat",
            typefname = "Type_Clust_res.dat",
        )
        visualis_pymol(
            water_type = water_types,
            waterO = coord_O,
            waterH1 = coord_H1,
            waterH2 = coord_H2,
            aligned_protein = "aligned.pdb",
            output_file = "results.pse",
            active_site_ids = [1,5,77,98],
            crystal_waters = "3t73",
            ligand_resname = "UBX",
            dist = 8.0,
        )
    """
    try:
        import pymol
        from pymol import cmd
    except ModuleNotFoundError:
        raise Exception("pymol not installed. Either install pymol or use nglview")
    if output_file is None and platform.system() != "Darwin" and lunch_pymol:
        pymol.finish_launching(["pymol", "-q"])
    if reinitialize:
        cmd.reinitialize()
    if platform.system() == "Darwin":
        import warnings

        warnings.warn(
            "mac OS detected interactive pymol session cannot be lunched. Visualisation state will be saved to pymol_water_visualization.pse",
            RuntimeWarning,
        )
        if output_file is None:
            output_file = "pymol_water_visualization.pse"
    cmd.hide("everything")
    if aligned_protein is not None:
        cmd.load(aligned_protein)
        cmd.hide("everything")
        # polymer for surface def
        tmpObj = cmd.get_unused_name("_tmp")
        cmd.create(tmpObj, "( all ) and polymer", zoom=0)
        # aminoacids in active site
        if active_site_ids:
            selection = ""
            for i in active_site_ids:
                selection += str(i) + "+"
            selection = selection[: len(selection) - 1]
            selection = "resi " + selection
            aminokis_u_am: str = cmd.get_unused_name("active_site_aa")
            cmd.select(
                aminokis_u_am,
                " " + selection + " and polymer ",
            )
            cmd.show("licorice", aminokis_u_am)
            cmd.color("magenta", aminokis_u_am)
            # pseudoatom in active site center
            active_site_center = cmd.get_unused_name("activesite_center_")
            cmd.pseudoatom(active_site_center, aminokis_u_am)
            cmd.hide(representation="everything", selection=active_site_center)
        # cmd.show("sphere",active_site_center)
        # cmd.set ("sphere_scale",0.1,active_site_center)
        # protein surface
        protein = cmd.get_unused_name("only_protein_")
        cmd.select(protein, "polymer")
        povrsina = cmd.get_unused_name("protein_surface_")
        cmd.create(povrsina, protein)
        cmd.show("surface", povrsina)
        cmd.color("gray70", povrsina)
        cmd.set("transparency", 0.5, povrsina)
        # ligand representation
        ligand = cmd.get_unused_name("ligand_")
        cmd.select(ligand, "organic")
        cmd.show("licorice", ligand)
    cntr = {"FCW": 0, "WCW": 0, "HCW": 0}
    for tip, Opos, H1pos, H2pos in zip(water_type, waterO, waterH1, waterH2):
        cntr[tip] += 1
        wname = tip + str(cntr[tip])
        print(wname)
        cmd.fetch("hoh", wname)
        cmd.alter_state(
            0,
            wname,
            "(x,y,z)=(" + str(Opos[0]) + "," + str(Opos[1]) + "," + str(Opos[2]) + ")",
        )
        if tip == "onlyO":
            cmd.delete(wname + "and elem H")
        else:
            indiciesH: list[int] = []
            cmd.iterate_state(
                -1,
                wname + " and elem H",
                "indiciesH.append(index)",
                space={"indiciesH": indiciesH},
            )
            cmd.alter_state(
                0,
                wname + " and index " + str(indiciesH[0]),
                "(x,y,z)=("
                + str(H1pos[0])
                + ","
                + str(H1pos[1])
                + ","
                + str(H1pos[2])
                + ")",
            )
            cmd.alter_state(
                0,
                wname + " and index " + str(indiciesH[1]),
                "(x,y,z)=("
                + str(H2pos[0])
                + ","
                + str(H2pos[1])
                + ","
                + str(H2pos[2])
                + ")",
            )
            if output_file is None or output_file.endswith(".pse"):
                cmd.alter_state(
                    0,
                    wname,
                    "resn='SOL'",
                )
                cmd.show("lines", wname)
                if tip == "FCW":
                    cmd.color("firebrick", wname)
                elif tip == "HCW":
                    cmd.color("skyblue", wname)
                elif tip == "WCW":
                    cmd.color("limegreen", wname)
            elif output_file.endswith(".pdb"):
                cmd.alter_state(
                    0,
                    wname,
                    "resn='" + str(tip) + "'",
                )

        if tip == "onlyO":
            cmd.show("spheres", wname)
            cmd.set("sphere_scale", 0.1, wname)
        else:
            cmd.show("sticks", wname)
    waters: str = cmd.get_unused_name("waters")
    cmd.select(waters, "SOL in (FCW* or WCW* or HCW*)")
    if active_site_ids:
        sele = aminokis_u_am + " or " + waters + " or organic"
    else:
        sele = waters + " or organic"
    cmd.distance(
        "polar_contacts",
        sele,
        "sol",
        mode=2,
    )
    cmd.hide("labels")
    # Add crystal waters
    if crystal_waters and aligned_protein is not None:
        cmd.fetch(crystal_waters)
        cmd.hide("everything", crystal_waters)
        cmd.align(
            "polymer and " + crystal_waters,
            "polymer and " + protein,
        )
        if ligand_resname is not None:
            cmd.select(
                "crystal_waters",
                "("
                + crystal_waters
                + " and SOL) within "
                + str(dist)
                + " of resname "
                + ligand_resname,
            )
        elif active_site_ids is not None:
            cmd.select(
                "crystal_waters",
                "("
                + crystal_waters
                + " and SOL) within 10 of resname "
                + active_site_center,
            )
        else:
            cmd.select("crystal_waters", crystal_waters + " and SOL")
        cmd.show("spheres", "crystal_waters")
        cmd.set("sphere_scale", "0.4", "crystal_waters")
        if os.path.exists(crystal_waters + ".cif"):
            os.remove(crystal_waters + ".cif")
    # add volume density visualisation
    if density_map:
        cmd.load(density_map)
        cmd.volume("water_density", density_map.split(".")[0])
    # reset camera
    cmd.reset()
    if active_site_ids is not None:
        cmd.center(active_site_center)
    if os.path.exists("hoh.cif"):
        os.remove("hoh.cif")
    # save
    if output_file is not None:
        cmd.save(output_file)


def visualise_pymol_from_pdb(
    pdbfile: str,
    active_site_ids: list[int] | None = None,
    crystal_waters: str | None = None,
    ligand_resname: str | None = None,
    dist: float = 10.0,
    density_map: str | None = None,
) -> None:
    """Make a `pymol <https://pymol.org/>`__ session from a pdb file.

    Visualises a pdb made by :py:meth:`make_results_pdb_MDA` file with water
    clustering results in pymol.

    Args:
        pdbfile (str): Name of the pdb file to read, should end in .pdb
        active_site_ids (list[int] | None, optional): Residue ids -
            numbers of aminoacids in active site. These are visualised
            as licorice. Defaults to None.
        crystal_waters (str | None, optional): PDBid from which crystal
            waters will attempted to be extracted. Defaults to None.
        ligand_resname (str | None, optional): Residue name of the
            ligand around which crystal waters (oxygens) shall be
            selected. Defaults to None.
        dist (float): distance from the centre of ligand around which
            crystal waters shall be selected. Defaults to 10.0.
        density_map (str | None, optional): Water density map to add to
            visualisation session (usually .dx file). Defaults to None.

    Example::

        visualise_pymol_from_pdb(
            pdbfile = "results.pdb",
            active_site_ids = [1,5,77,98],
            crystal_waters = "3t73",
            ligand_resname = "UBX",
            dist = 8.0,
            density_map = "waters.dx"
        )
    """
    try:
        import pymol
        from pymol import cmd
    except ModuleNotFoundError:
        raise Exception("pymol not installed. Either install pymol or use nglview")
    if platform.system() != "Darwin":
        pymol.finish_launching(["pymol", "-q"])
    cmd.load(pdbfile)
    cmd.hide("everything")
    # polymer for surface def
    tmpObj = cmd.get_unused_name("_tmp")
    cmd.create(tmpObj, "( all ) and polymer", zoom=0)
    # aminoacids in active site
    if active_site_ids:
        aminokis_u_am = cmd.get_unused_name("active_site_aa")
        selection = ""
        for i in active_site_ids:
            selection += str(i) + "+"
        selection = selection[: len(selection) - 1]
        selection = "resi " + selection
        cmd.select(
            aminokis_u_am,
            " " + selection + " and polymer ",
        )
        cmd.show("licorice", aminokis_u_am)
        cmd.color("magenta", aminokis_u_am)
        # pseudoatom in active site center
        active_site_center = cmd.get_unused_name("activesite_center_")
        cmd.pseudoatom(active_site_center, aminokis_u_am)
        cmd.hide(representation="everything", selection=active_site_center)
    # cmd.show("sphere",active_site_center)
    # cmd.set ("sphere_scale",0.1,active_site_center)
    # protein surface
    protein = cmd.get_unused_name("only_protein_")
    cmd.select(protein, "polymer")
    povrsina = cmd.get_unused_name("protein_surface_")
    cmd.create(povrsina, protein)
    cmd.show("surface", povrsina)
    cmd.color("gray70", povrsina)
    cmd.set("transparency", 0.5, povrsina)
    # ligand representation
    ligand = cmd.get_unused_name("ligand_")
    cmd.select(ligand, "organic")
    cmd.show("licorice", ligand)
    # add water representations
    # conserved waters
    conserved = cmd.get_unused_name("FCW_")
    cmd.select(conserved, "resname FCW")
    cmd.show("lines", conserved)
    cmd.color("firebrick", conserved)
    # half conserved waters
    half_conserved = cmd.get_unused_name("HCW_")
    cmd.select(half_conserved, "resname HCW")
    cmd.show("lines", half_conserved)
    cmd.color("skyblue", half_conserved)
    # semi conserved waters
    semi_conserved = cmd.get_unused_name("WCW_")
    cmd.select(semi_conserved, "resname WCW")
    cmd.show("lines", semi_conserved)
    cmd.color("limegreen", semi_conserved)
    # all waters
    waters = cmd.get_unused_name("waters_")
    cmd.select(
        waters,
        semi_conserved + " or " + half_conserved + " or " + conserved,
    )
    # Add crystal waters
    if crystal_waters:
        cmd.fetch(crystal_waters)
        cmd.hide("everything", crystal_waters)
        cmd.align(
            "polymer and " + crystal_waters,
            "polymer and " + pdbfile.split(".")[0],
        )
        if ligand_resname:
            cmd.select(
                "crystal_waters",
                "("
                + crystal_waters
                + " and SOL) within "
                + str(dist)
                + " of resname "
                + ligand_resname,
            )
        else:
            cmd.select("crystal_waters", crystal_waters + " and SOL")
        cmd.show("spheres", "crystal_waters")
        cmd.set("sphere_scale", "0.4", "crystal_waters")
    # add volume density visualisation
    if density_map:
        cmd.load(density_map)
        cmd.volume("water_density", density_map.split(".")[0])
    # reset camera
    cmd.reset()
    if active_site_center is not None:
        cmd.center(active_site_center)
    # save
    cmd.reinitialize()


def visualise_nglview(
    water_type: list[str],
    waterO: list[list[float]],
    waterH1: list[list[float]],
    waterH2: list[list[float]],
    aligned_protein: str = "aligned.pdb",
    active_site_ids: list[int] | None = None,
    crystal_waters: str | None = None,
    density_map_file: str | None = None,
) -> NGLWidget:
    """Creates `nglview <https://github.com/nglviewer/nglview>`__  visualisation widget for results.

    Starts a nglview visualisation instance from clustering results.

    Args:
        water_type (list): List containing water type results from
            water clustering.
        waterO (list): Coordiantes of Oxygen atom in water molecules.
        waterH1 (list): Coordinates of Hydrogen1 atom in water molecules.
        waterH2 (list): Coordinates of Hydrogen2 atom in water molecules.
        aligned_protein (str, optional): file name containing protein
            configuration trajectory was aligned to. Defaults to "aligned.pdb".
        active_site_ids (list[int] | None, optional): Residue ids -
            numbers of aminoacids in active site. These are visualised
            as licorice. Defaults to None.
        crystal_waters (str | None, optional): PDBid from which crystal
            waters will attempted to be extracted. Defaults to None.
        density_map (str | None, optional): Water density map to add to
            visualisation session (usually .dx file). Defaults to None.

    Returns:
        NGLWidget: Returns nglview Widget for visualisation of the
        results.

    Example::

        # read results and visualise them using nglview
        water_types, coord_O, coord_H1, coord_H2 = read_results(
            fname = "Clust_res.dat",
            typefname = "Type_Clust_res.dat",
        )
        view = visualise_nglview(
            water_type = water_types,
            waterO = coord_O,
            waterH1 = coord_H1,
            waterH2 = coord_H2,
            aligned_protein = "aligned.pdb",
            active_site_ids = [1,5,77,98],
            crystal_waters = "3t73",
        )
        # initialise widget
        view
    """
    try:
        import nglview as ngl
    except ModuleNotFoundError:
        raise Exception("nglview not installed. Either install pymol or nglview")

    if aligned_protein is not None:
        view: NGLWidget = ngl.show_file(aligned_protein, default_representation=False)
        view.clear_representations()
        view.add_representation("surface", selection="protein", opacity=0.5)
        view.add_representation("ball+stick", selection="water", color="red")
        selection = ""
        if active_site_ids is not None:
            for i in active_site_ids:
                selection = selection[: len(selection) - 3]
                selection += str(i) + " or "
            view.add_representation(
                "ball+stick",
                selection=selection,
                color="pink",
            )
    else:
        view = ngl.NGLWidget()
    col = {"FCW": "red", "WCW": "blue", "HCW": "green"}
    for tip, Opos, H1pos, H2pos in zip(water_type, waterO, waterH1, waterH2):
        view.add_pdbid("hoh")
        view[-1].add_representation("licorice", color=col[tip])
        view[-1].set_coordinates(np.asarray([H1pos, Opos, H2pos]))
    if crystal_waters is not None:
        view.add_pdbid(crystal_waters)
        view[-1].add_representation("spacefill", selection="water", color="red")
    if density_map_file is not None:
        view.add_component(density_map_file)
    return view



"""
Unit and regression test for the ConservedWaterSearch package.
"""

# Import package, test suite, and other packages as needed
import ConservedWaterSearch.hydrogen_orientation
import pytest
import numpy as np

make_ho_plots = 0


def test_orientation_normalization():
    orientations = np.asarray([[1, 0, 0], [-0.25038 * 2, 0.96814764 * 2, 0]])
    orientations2 = np.asarray([[1, 0, 0], [-0.25038, 0.96814764, 0]])
    w1 = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations, normalize_orientations=True
    )
    w2 = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations2
    )
    print(w1, w2)
    assert w1 == w2


def test_orientation_shape():
    orientations = np.asarray([[[1]]])
    with pytest.raises(ValueError):
        ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
            orientations
        )


def test_orientation_dimensions():
    orientations = np.asarray([[1, 0], [0, 1]])
    with pytest.raises(ValueError):
        ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
            orientations
        )


def test_orientation_valid_input():
    orientations = np.asarray([[1, 0, 0], [0, 1, 0]])
    ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations
    )


def test_orientation_size():
    orientations = np.asarray([[1, 0, 0]])
    with pytest.raises(ValueError):
        ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
            orientations
        )


def test_orientation_array_odd():
    orientations = np.asarray([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    with pytest.raises(ValueError):
        ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
            orientations
        )


def test_conserved_FCW():
    orientations = np.loadtxt("tests/data/conserved_sample_FCW.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "FCW"


def test_conserved_FCW2():
    orientations = np.loadtxt("tests/data/conserved_sample_FCW2.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "FCW"


def test_dispersed_HCW():
    orientations = np.loadtxt("tests/data/dispersed_sample_HCW.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "HCW"


def test_dispersed_HCW2():
    orientations = np.loadtxt("tests/data/dispersed_sample_HCW2.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "HCW"


def test_circular_HCW():
    orientations = np.loadtxt("tests/data/circ_sample_HCW.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "HCW"


def test_circular_HCW2():
    orientations = np.loadtxt("tests/data/circ_sample_HCW2.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "HCW"


def test_circular_HCW3():
    orientations = np.loadtxt("tests/data/circ_sample_HCW3.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "HCW"


def test_WCW():
    orientations = np.loadtxt("tests/data/sample_WCW.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "WCW"


def test_circular_only():
    orientations = np.loadtxt("tests/data/sample_circular_only.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "WCW"


def test_dispersed_WCW():
    orientations = np.loadtxt("tests/data/dispersed_sample_WCW.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) > 0
    assert res[0][2] == "WCW"


def test_2x2_WCW():
    orientations = np.loadtxt("tests/data/2_by_2_WCW.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) == 2
    assert res[0][2] == "WCW"


def test_not_conserved():
    orientations = np.loadtxt("tests/data/dispersed_sample_not_conserved.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) == 0


def test_not_conserved2():
    orientations = np.loadtxt("tests/data/not_conserved.dat")
    res = ConservedWaterSearch.hydrogen_orientation.hydrogen_orientation_analysis(
        orientations,
        debugH=make_ho_plots,
        verbose=2,
    )
    if make_ho_plots > 0:
        import matplotlib.pyplot as plt

        plt.show()
    assert len(res) == 0



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    try:
        from matplotlib.axes import Axes
        from matplotlib.figure import Figure
    except ImportError:
        Axes, Figure = None, None

import numpy as np
from sklearn.cluster import OPTICS, KMeans


def hydrogen_orientation_analysis(
    orientations: np.ndarray,
    pct_size_buffer: float = 0.85,
    kmeans_ang_cutoff: float = 120,
    kmeans_inertia_cutoff: float = 0.4,
    FCW_angdiff_cutoff: float = 5,
    FCW_angstd_cutoff: float = 17,
    min_samp_data_size_pct: float = 0.15,
    nonFCW_angdiff_cutoff: float = 15,
    HCW_angstd_cutoff: float = 17,
    WCW_angstd_cutoff: float = 20,
    weakly_explained: float = 0.7,
    xiFCW: list[float] = [0.03],
    xiHCW: list[float] = [0.05, 0.01],
    xiWCW: list[float] = [0.05, 0.001],
    njobs: int = 1,
    verbose: int = 0,
    debugH: int = 0,
    plotreach: bool = False,
    which: list[str] = ["FCW", "HCW", "WCW"],
    normalize_orientations: bool = True,
) -> list:
    """Determines if the water cluster is conserved and of what type.

    High level function that does hydrogen orientation analysis. Checks
    if the water cluster belongs into one of the following groups by
    analizing hydrogen orientations:

    - FCW (Fully Conserved Water): hydrogens are strongly oriented in
      two directions with angle of 104.5
    - HCW (Half Conserved Water): one set (cluster) of hydrogens is
      oriented in certain directions and other are spread into different
      orientations with angle of 104.5
    - WCW (Weakly Conserved Water): several orientation combinations
      exsist with satisfying water angles

    See :cite:`conservedwatersearch2022` for more information on water
    classification :ref:`conservedwaters:theory, background and methods`.
    If orientations don't satisfy the criteria for any of the waters, an
    empty list is returned.

    Args:
        orientations (np.ndarray): array of hydrogen
            orientations in space
        pct_size_buffer (float, optional): Minimum allowed size of the
            hydrogen orientation cluster. Defaults to 0.85.
        kmeans_ang_cutoff (float, optional): Maximum value of angle (in
            deg) allowed for FCW in kmeans clustering to be considered
            correct water angle. Defaults to 120.
        kmeans_inertia_cutoff (float, optional): upper limit allowed on
            kmeans inertia (measure of spread of data in a cluster).
            Defaults to 0.4.
        FCW_angdiff_cutoff (float, optional): Maximum value of angle (in
            deg) allowed for FCW in OPTICS/HDBSCAN clustering to be
            considered correct water angle. Defaults to 5.
        FCW_angstd_cutoff (float, optional): Maximal standard deviation
            of angle distribution of orientations of two hydrogens
            allowed for water to be considered FCW. Defaults to 17.
        min_samp_data_size_pct (float, optional): Minimum samples to
            choose for OPTICS or HDBSCAN clustering as percentage of
            number of water molecules considered for HCW and WCW.
            Defaults to 0.15.
        nonFCW_angdiff_cutoff (float, optional): Maximum standard
            deviation of angle allowed for HCW and WCW to be considered
            correct water angle. Defaults to 15.
        HCW_angstd_cutoff (float, optional): Maximum standard deviation
            cutoff for WCW angles to be considered correct water angles.
            Defaults to 17.
        WCW_angstd_cutoff (float, optional): Maximum standard deviation
            cutoff for WCW angles to be considered correct water angles.
            Defaults to 20.
        weakly_explained (float, optional): percentage of explained
            hydrogen orientations for water to be considered WCW.
            Defaults to 0.7.
        xiFCW (list, optional): Xi value for OPTICS clustering for FCW. Don't
            touch this unless you know what you are doing. Defaults to [0.03].
        xiHCW (list, optional): Xi value for OPTICS clustering for HCW. Don't
            touch this unless you know what you are doing.
            Defaults to [0.05, 0.01].
        xiWCW (list, optional): Xi value for OPTICS clustering for WCW. Don't
            touch this unless you know what you are doing.
            Defaults to [0.05, 0.001].
        njobs (int, optional): how many cpu cores to use for clustering.
            Defaults to 1.
        verbose (int, optional): verbosity of output. Defaults to 0.
        debugH (int, optional): debug level for orientations. Defaults to 0.
        plotreach (bool, optional): weather to plot the reachability
            plot for OPTICS when debuging. Defaults to False.
        which (list, optional): list of strings denoting which water types to
            search for. Allowed is any combination of FCW (fully
            conserved waters), HCW (half conserved waters) and WCW
            (weakly conserved waters). Defaults to ["FCW", "HCW",
            "WCW"].
        normalize_orientations (bool, optional): weather to normalize
            the orientation vectors to unit distance. Defaults to True.

    Returns:
        list: returns a list containing two orientations of hydrogens
        and water classification string ("FCW", "HCW", "WCW"), if
        not conserved returns an empty list
    """
    orientations = np.array(orientations)
    # check length of orientations - it has to be bigger then 1 and even
    orishape = orientations.shape
    if len(orishape) != 2:
        raise ValueError("Orientations have to be a 2D array")
    Hnum = orishape[0]
    dimnum = orishape[1]
    if dimnum != 3:
        raise ValueError("Orientations must be vectors of dimension 3")
    if Hnum < 2 or Hnum % 2 != 0:
        raise ValueError(
            "Number of orientations must be even! Each water molecule has 2 hydrogen atoms!"
        )
    if normalize_orientations:
        orientations = orientations / np.linalg.norm(
            orientations, axis=1, keepdims=True
        )
    # For Debug
    np.set_printoptions(precision=4)
    # Check if water is conserved
    if "FCW" in which:
        for xi in xiFCW:
            conserved = find_fully_conserved_orientations(
                orientations,
                pct_size_buffer=pct_size_buffer,
                kmeans_ang_cutoff=kmeans_ang_cutoff,
                kmeans_inertia_cutoff=kmeans_inertia_cutoff,
                angdiff_cutoff=FCW_angdiff_cutoff,
                angstd_cutoff=FCW_angstd_cutoff,
                xi=xi,
                njobs=njobs,
                verbose=verbose,
                debugH=debugH,
                plotreach=plotreach,
            )
            if len(conserved) > 0:
                return conserved
    if "HCW" in which:
        for xi in xiHCW:
            half_conserved = find_half_conserved_orientations(
                orientations,
                pct_size_buffer=pct_size_buffer,
                min_samp_data_size_pct=min_samp_data_size_pct,
                angdiff_cutoff=nonFCW_angdiff_cutoff,
                angstd_cutoff=HCW_angstd_cutoff,
                xi=xi,
                njobs=njobs,
                verbose=verbose,
                debugH=debugH,
                plotreach=plotreach,
            )
            if len(half_conserved) > 0:
                return half_conserved
    if "WCW" in which:
        for xi in xiWCW:
            semi_conserved = find_weakly_conserved_orientations(
                orientations,
                pct_size_buffer=pct_size_buffer,
                min_samp_data_size_pct=min_samp_data_size_pct,
                pct_explained=weakly_explained,
                angdiff_cutoff=nonFCW_angdiff_cutoff,
                angstd_cutoff=WCW_angstd_cutoff,
                xi=xi,
                njobs=njobs,
                verbose=verbose,
                debugH=debugH,
                plotreach=plotreach,
            )
            if len(semi_conserved):
                return semi_conserved
    return []
    # I want this to return an array with hydrogen orientations combinations


def find_fully_conserved_orientations(
    orientations: np.ndarray,
    pct_size_buffer: float = 0.85,
    kmeans_ang_cutoff: float = 120,
    kmeans_inertia_cutoff: float = 0.4,
    angdiff_cutoff: float = 5,
    angstd_cutoff: float = 17.0,
    xi: float = 0.03,
    njobs: int = 1,
    verbose: int = 0,
    debugH: int = 0,
    plotreach: bool = False,
) -> list:
    """Check if orientations belong to FCW.

    Checks if given oxygen cluster can be considered as a fully
    conserved water based on hydrogen orientations. Fully conserved
    water is one which has well defined hydrogen orientations in two
    distinctive groups (ie strongly hydrogen bonded for both hydrogens).
    To check if water is conserved, one first checks if k means
    clustering of hydrogen orientations gives two destinctive clusters
    with low inertia and required angle between the clusters. Afterwards
    more rigorous check is carried out with OPTICS clustering where
    again the sperad of orientations and angle is considered.

    Args:
        orientations (np.ndarray): array of hydrogen
            orientations in space
        pct_size_buffer (float, optional): Minimum allowed size of the
            hydrogen orientation cluster. Defaults to 0.85.
        kmeans_ang_cutoff (float, optional): Maximum value of angle (in
            deg) allowed for FCW in kmeans clustering to be considered
            correct water angle. Defaults to 120.
        kmeans_inertia_cutoff (float, optional): upper limit allowed on
            kmeans inertia (measure of spread of data in a cluster).
            Defaults to 0.4.
        angdiff_cutoff (float, optional): Maximum value of angle (in
            deg) allowed for FCW in OPTICS/HDBSCAN clustering to be
            considered correct water angle. Defaults to 5.
        angstd_cutoff (float, optional): Maximal standard deviation
            of angle distribution of orientations of two hydrogens
            allowed for water to be considered FCW. Defaults to 17.
        xi (float, optional): Xi value for OPTICS clustering for FCW. Don't
            touch this unless you know what you are doing.
            Defaults to 0.03.
        njobs (int, optional): how many cpu cores to use for clustering.
            Defaults to 1.
        verbose (int, optional): verbosity of output. Defaults to 0.
        debugH (int, optional): debug level for orientations. Defaults to 0.
        plotreach (bool, optional): weather to plot the reachability
            plot for OPTICS when debuging. Defaults to False.

    Returns:
        list: returns a list containing two orientations of hydrogens
        and water classification string "FCW", if not FCW returns
        empty list
    """
    # number of elements in oxygen cluster
    neioc = int(len(orientations) / 2)
    # for debugging
    counts = np.asarray([0])
    values = np.asarray([0])
    angs12 = 0
    angs21 = 0
    biggest = 0
    secondbiggest = 0
    labels = 0
    cc = 0
    ss = ""
    # conserved hydrogens
    fully_conserved = []
    # Kmeans clustering prepareation - for conserved water analysis only
    kmeans = KMeans(n_clusters=2, n_init=10)
    # fit kmeans clustering
    kmeans.fit(orientations)
    if len(kmeans.cluster_centers_) == 2:
        Kcv1 = kmeans.cluster_centers_[0]
        Kcv2 = kmeans.cluster_centers_[1]
        # Calculate angles between centers of clusters that represent hydrogen orientation
        ang = (
            360.0
            / (2.0 * np.pi)
            * np.arccos(
                np.clip(
                    np.dot(
                        Kcv1 / np.linalg.norm(Kcv1),
                        Kcv2 / np.linalg.norm(Kcv2),
                    ),
                    -1.0,
                    1.0,
                )
            )
        )
        # check if angle between cluster centres is about right and check if spread (variance) of orientations is sufficiently low
        if (
            ang < kmeans_ang_cutoff
            and kmeans.inertia_ / len(orientations) < kmeans_inertia_cutoff
        ):
            # Perform OPTICS clustering on hydrogen orientations; minsamp is same for all water types, however here we force min_cluster size to be given.
            msp = int(neioc * pct_size_buffer)
            msp = msp if msp >= 2 else 2
            cc = OPTICS(min_samples=msp, xi=xi, n_jobs=njobs)
            cc.fit(orientations)
            labels = cc.labels_
            # Calculate number of elements in each cluster
            (values, counts) = np.unique(labels[labels != -1], return_counts=True)
            # if number of optics clusters for hydrogen clustering is >0
            if len(np.sort(np.unique(labels[labels != -1]))) == 2:
                # find two biggest clusters
                biggest = values[np.argsort(counts)[-1]]
                secondbiggest = values[np.argsort(counts)[-2]]
                # check if size of two biggest clusters are between 0.85 and 1.15*sizeofOxygenCluster
                if (
                    len(orientations[labels == biggest])
                    <= neioc * (2 - pct_size_buffer)
                    and len(orientations[labels == secondbiggest])
                    <= neioc * (2 - pct_size_buffer)
                    and len(orientations[labels == biggest]) > neioc * pct_size_buffer
                    and len(orientations[labels == secondbiggest])
                    > neioc * pct_size_buffer
                ):
                    # calculate the average orientation of two biggest clusters
                    avang1 = np.mean(orientations[labels == biggest], axis=0)
                    avang2 = np.mean(orientations[labels == secondbiggest], axis=0)
                    # calculate average angle between average orientation of one cluster with elements(orientations) of the other cluster
                    angs12 = (
                        np.arccos(
                            np.dot(orientations[labels == secondbiggest], avang1.T)
                        )
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    angs21 = (
                        np.arccos(np.dot(orientations[labels == biggest], avang2.T))
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    # check if average angles are low and that their std. deviation is low
                    if (
                        np.abs(np.mean(angs21) - np.mean(angs12)) < angdiff_cutoff
                        and np.std(angs12) < angstd_cutoff
                        and np.std(angs21) < angstd_cutoff
                    ):
                        # conserved
                        if verbose > 0:
                            print("conserved found")
                        fully_conserved.append(
                            __return_normalized_orientation_pair(
                                orientations,
                                labels,
                                biggest,
                                secondbiggest,
                                "FCW",
                            )
                        )
    # Debug needs neoic,orientations,kmeans,ang,k,counts,biggest,secondbiggest,angs12,angs21,cc
    # kmeans
    if verbose > 0 or debugH > 0:
        sk = (
            f"Conserved - kmeans analysis:\n"
            f"angle - mean: {ang:.2f}(calced)<{kmeans_ang_cutoff:.2f}; inertia per H: {kmeans.inertia_/len(orientations):.2f}(calced)<{kmeans_inertia_cutoff:.2f}\n"
        )
    if debugH == 2 or (debugH == 1 and len(fully_conserved) > 0):
        __plot3Dorients(111, kmeans.labels_, orientations, sk)
    # OPTICS
    if debugH > 0 or verbose > 0:
        ss = (
            f"Fully Conserved - OPTICS analysis;depends on pct size buffer:{pct_size_buffer:.2f}\n"
            f"cluster sizes: {counts}; clust labels:{values};biggest: {biggest}; secondbiggest: {secondbiggest}\n"
            f"clusters size: {neioc*pct_size_buffer:.2f}<{len(orientations[labels == biggest])},{len(orientations[labels == secondbiggest])}(calced)<{neioc*(2-pct_size_buffer):.2f}\n"
            f"angle mean: {np.abs(np.mean(angs21)-np.mean(angs12)):.2f}(calced)<{angdiff_cutoff:.2f}); 12 std : {np.std(angs12):.2f};21 std: {np.std(angs21):.2f} (both <{angstd_cutoff:.2f})\n"
        )
    # Printing
    if verbose == 2 or (verbose == 1 and len(fully_conserved) > 0):
        print(sk + ss)
    # Debug plots
    __hydrogen_orient_plots(
        labels,
        orientations,
        cc,
        ss,
        f"Reachability of optics plot\n minsamples={int(neioc*pct_size_buffer)}; xi={xi}",
        len(fully_conserved) > 0,
        debugH,
        plotreach,
    )
    # End of function
    return fully_conserved


def find_half_conserved_orientations(
    orientations: np.ndarray,
    pct_size_buffer: float = 0.85,
    min_samp_data_size_pct: float = 0.35,
    angdiff_cutoff: float = 15,
    angstd_cutoff: float = 17.0,
    xi: float = 0.01,
    njobs: int = 1,
    verbose: int = 0,
    debugH: int = 0,
    plotreach: bool = False,
) -> list:
    """Checks if given orientations belong to HCW.

    Checks if given oxygen cluster can be considered as a half conserved
    water based on hydrogen orientations. Half conserved water is one
    which has one well defined hydrogen orientation (ie one strongly
    hydrogen bonded hydrogen). To check if water is half conserved, one
    calculates OPTICS clustering of hydrogen orientations. One then
    loops over clusters in an atempt to find a hydrogen orientation
    cluster which is the size of oxygen cluster and weather the angle
    between that cluster with all other orientations is of right angle
    and if spread of orientations is sufficiently low.

    Args:
        orientations (np.ndarray): array of hydrogen
            orientations in space
        pct_size_buffer (float, optional): Minimum allowed size of the
            hydrogen orientation cluster. Defaults to 0.85.
        min_samp_data_size_pct (float, optional): Minimum samples to
            choose for OPTICS or HDBSCAN clustering as percentage of
            number of water molecules considered for HCW and WCW.
            Defaults to 0.15.
        angdiff_cutoff (float, optional): Maximum standard
            deviation of angle allowed for HCW to be considered
            correct water angle. Defaults to 15.
        HCW_angstd_cutoff (float, optional): Maximum standard deviation
            cutoff for WCW angles to be considered correct water angles.
            Defaults to 17.
        xi (float, optional): Xi value for OPTICS clustering for HCW. Don't
            touch this unless you know what you are doing.
            Defaults to 0.01.
        njobs (int, optional): how many cpu cores to use for clustering.
            Defaults to 1.
        verbose (int, optional): verbosity of output. Defaults to 0.
        debugH (int, optional): debug level for orientations. Defaults to 0.
        plotreach (bool, optional): weather to plot the reachability
            plot for OPTICS when debuging. Defaults to False.

    Returns:
        list: returns a list containing two orientations of hydrogens
        and water classification string "HCW", if not HCW returns
        an empty list
    """
    # number of elements in oxygen cluster
    neioc = int(len(orientations) / 2)
    half_conserved = []
    sk = ""
    ss = ""
    # Optics clustering for hydrogen orientations - minsamp is same for all water types
    msp = int(neioc * min_samp_data_size_pct)
    msp = msp if msp >= 2 else 2
    cc = OPTICS(
        min_samples=msp,
        xi=xi,
        n_jobs=njobs,
    )
    cc.fit(orientations)
    # lebels without -1
    labels = cc.labels_[cc.labels_ != -1]
    (values, counts) = np.unique(labels, return_counts=True)
    ind = np.argmax(counts[counts != -1])
    bestcand = values[ind]
    labels = cc.labels_
    # if there is more then 1 cluster
    if len(np.unique(labels)) > 1:
        N_elems = len(orientations[labels == bestcand])
        # Debug
        if verbose > 0:
            sk += f"Cluster {bestcand} has {neioc*pct_size_buffer:.2f}<{N_elems}<{neioc*(2-pct_size_buffer):.2f} elements\n"
        # check if number of elements in hydrogen orientation cluster is between 0.80 and 1.20 elements of Oxygen cluster
        if (
            N_elems < neioc * (2 - pct_size_buffer)
            and N_elems > neioc * pct_size_buffer
        ):
            # calculating average angle between average orientation of selected cluster and all other orientations
            avang1 = np.mean(orientations[labels == bestcand], axis=0)
            angs1all = (
                np.arccos(np.dot(orientations[labels != bestcand], avang1.T))
                * 360.0
                / (2.0 * np.pi)
            )
            # Debug
            if verbose > 0 or debugH > 0:
                sk += f"result angles: mean= {np.abs(np.mean(angs1all)-104.5):.2f}(Calced)<{angdiff_cutoff:.2f}; std dev={np.std(angs1all):.2f}(Calced)<{angstd_cutoff:.2f}\n"
            # check if angle and its std deviation is satisfactory
            if (
                np.abs(np.mean(angs1all) - 104.5) < angdiff_cutoff
                and np.std(angs1all) < angstd_cutoff
            ):
                for j in np.unique(labels[labels != bestcand]):
                    # -1 has to be inclueded in case main cluster is ok and -1 is everywhere, to produce a HCW still
                    if verbose > 0:
                        print("half conserved found", bestcand, j)
                    half_conserved.append(
                        __return_normalized_orientation_pair(
                            orientations, labels, bestcand, j, "HCW"
                        )
                    )
    # Debug
    if verbose > 0 or debugH > 0:
        ss = (
            f"Half Conserved - OPTICS analysis;xi={xi}"
            f"best: {bestcand}; N_hyd_cls (w/o -1): {len(np.unique(labels[labels!=-1]))};\n N elem in biggest:{neioc*(2-pct_size_buffer):.2f}>{len(orientations[labels==bestcand])}>{neioc*pct_size_buffer:.2f}\n"
        )
    # Printing
    if verbose == 2 or (verbose == 1 and len(half_conserved) > 0):
        print(ss + sk)
    # Debug plots
    __hydrogen_orient_plots(
        labels,
        orientations,
        cc,
        ss + sk,
        f"Reachability of optics plot\n minsamples={int(neioc*min_samp_data_size_pct)}; xi={xi}",
        len(half_conserved) > 0,
        debugH,
        plotreach,
    )
    # End of function
    return half_conserved


def find_weakly_conserved_orientations(
    orientations: np.ndarray,
    pct_size_buffer: float = 0.85,
    lower_bound_pct_buffer: float = 0.35,
    min_samp_data_size_pct: float = 0.15,
    pct_explained: float = 0.7,
    angdiff_cutoff: float = 15,
    angstd_cutoff: float = 20.0,
    xi: float = 0.01,
    njobs: int = 1,
    verbose: int = 0,
    debugH: int = 0,
    plotreach: bool = False,
) -> list:
    """Checks if given orientations belong to WCW.

    Checks if given oxygen cluster can be considered as a weakly
    conserved water based on hydrogen orientations. weakly conserved
    water is one which has no well defined hydrogen orientation (ie no
    strongly hydrogen bonded hydrogen) but still has distinct hydrogen
    orientational clusters. To check if water is weakly conserved, one
    calculates OPTICS clustering of hydrogen orientations. One then
    loops over clusters in an atempt to find a pair of hydrogen
    orientation clusters which is of the same size and weather the angle
    between the two clusters is of right angle and if spread of
    orientations is sufficiently low. Aditionally triplets are checked
    as well. Here we do the same check but we are looking at cluster one
    vs two other clusters combined.

    Args:
        orientations (np.ndarray): array of hydrogen
            orientations in space
        pct_size_buffer (float, optional): Minimum allowed size of the
            hydrogen orientation cluster. Defaults to 0.85.
        min_samp_data_size_pct (float, optional): Minimum samples to
            choose for OPTICS or HDBSCAN clustering as percentage of
            number of water molecules considered for HCW and WCW.
            Defaults to 0.15.
        pct_explained (float, optional): percentage of explained
            hydrogen orientations for water to be considered WCW.
            Defaults to 0.7.
        angdiff_cutoff (float, optional): Maximum standard
            deviation of angle allowed for WCW to be considered
            correct water angle. Defaults to 15.
        angstd_cutoff (float, optional): Maximum standard deviation
            cutoff for WCW angles to be considered correct water angles.
            Defaults to 20.
        xi (float, optional): Xi value for OPTICS clustering for WCW. Don't
            touch this unless you know what you are doing.
            Defaults to 0.01.
        njobs (int, optional): how many cpu cores to use for clustering.
            Defaults to 1.
        verbose (int, optional): verbosity of output. Defaults to 0.
        debugH (int, optional): debug level for orientations. Defaults to 0.
        plotreach (bool, optional): weather to plot the reachability
            plot for OPTICS when debuging. Defaults to False.

    Returns:
        list: returns a list containing two orientations of hydrogens
        and water classification string "WCW", if not WCW returns
        an empty list
    """
    # number of elements in oxygen cluster
    neioc = int(len(orientations) / 2)
    # Optics clustering for hydrogen orientations - minsamp is same for all water types
    msp = int(neioc * min_samp_data_size_pct)
    msp = msp if msp >= 2 else 2
    cc = OPTICS(
        min_samples=msp,
        xi=xi,
        n_jobs=njobs,
    )
    cc.fit(orientations)
    weakly_conserved = []
    # lebels without -1
    labels = cc.labels_
    (_, counts) = np.unique(labels, return_counts=True)
    weakly = []
    sk = ""
    ss = ""
    # array of already used clusters
    used = []
    # list of labels
    lbls = np.unique(labels)
    # if there is more then 1 cluster
    if len(np.unique(labels)) > 1:
        # Chekc over pairs of clusters
        # loop over OPTICS clusters
        for ii in lbls[lbls != -1]:
            # check if in used
            if ii in used:
                continue
            N_elems = len(orientations[labels == ii])
            # Debug
            if verbose > 0 or debugH > 0:
                sk += f"Cluster {ii} has {neioc*lower_bound_pct_buffer:.2f}<{N_elems}<{neioc*(2-pct_size_buffer):.2f} elements\n"
            # check if size of hydorgen orientation cluster is between 1.20 and lower_bound_pct_buffer times number of elements in oxygen cluster
            if (
                N_elems < neioc * (2 - pct_size_buffer)
                and N_elems > neioc * lower_bound_pct_buffer
            ):
                avang1 = np.mean(orientations[labels == ii], axis=0)
                # loop over clusters but not ii
                lblsni = lbls[lbls != ii]
                for jj in lblsni:
                    # check if already used
                    if ii in used:
                        break
                    if jj in used:
                        continue
                    # calculate average angle between average orientation of ii and all orientations in cluster jj
                    angs1j = (
                        np.arccos(np.dot(orientations[labels == jj], avang1.T))
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    N_elems_jj = len(orientations[labels == jj])
                    # Debug
                    if verbose > 0 or debugH > 0:
                        sk += f"cluster combo:{ii} & {jj}size:{N_elems},{neioc*lower_bound_pct_buffer:.2f}<{N_elems_jj}<{neioc*(2-pct_size_buffer):.2f}\n"
                        sk += f"size comparison {np.abs(N_elems -N_elems_jj)} (calced) < {(1 - pct_size_buffer) * np.max([N_elems, N_elems_jj])} \n"
                        sk += f"ang diff={np.abs(np.mean(angs1j)-104.5):.2f}(calced)<{angdiff_cutoff:.2f},std dev:{angstd_cutoff:.2f}>{np.std(angs1j):.2f}(calced)\n"
                    # check if size of new cluster and check if size of clusters is about equal
                    if (
                        N_elems_jj > neioc * (2 - pct_size_buffer)
                        or N_elems_jj < neioc * lower_bound_pct_buffer
                        or np.abs(N_elems - N_elems_jj)
                        > (1 - pct_size_buffer) * np.max([N_elems, N_elems_jj])
                    ):
                        continue
                    # check if angles and std devs are good
                    if (
                        np.abs(np.mean(angs1j) - 104.5) < angdiff_cutoff
                        and np.std(angs1j) < angstd_cutoff
                        and len(orientations[labels == jj])
                        > neioc * lower_bound_pct_buffer
                        and len(orientations[labels == jj])
                        < neioc * (2 - pct_size_buffer)
                    ):
                        used.append(ii)
                        used.append(jj)
                        weakly.append([ii, jj])
                        break
        # check for triplets if cluster ii has same size as clusters jj+kk and satisfactory angles
        # loop over OPTICS clusters
        for ii in lbls[lbls != -1]:
            # check if in used
            if ii in used:
                continue
            N_elems = len(orientations[labels == ii])
            # average orientation of ii
            avang1 = np.mean(orientations[labels == ii], axis=0)
            # loop over clusters but not ii
            lblsni = lbls[lbls != ii]
            for jj in lblsni:
                # check if already used
                if ii in used:
                    break
                if jj in used:
                    continue
                N_elems_jj = len(orientations[labels == jj])
                avangj = np.mean(orientations[labels == jj], axis=0)
                # loop over clusters but not ii or jj
                lblsnij = lblsni[lblsni != jj]
                for kk in lblsnij:
                    if ii in used:
                        break
                    if jj in used:
                        break
                    if kk in used:
                        continue
                    N_elems_kk = len(orientations[labels == kk])
                    avangk = np.mean(orientations[labels == kk], axis=0)
                    # calculate average angle between average orientation of ii and all orientations in cluster jj and kk
                    angs1jk = (
                        np.arccos(
                            np.dot(
                                orientations[(labels == jj) | (labels == kk)],
                                avang1.T,
                            )
                        )
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    angs1j = (
                        np.arccos(
                            np.dot(
                                orientations[(labels == jj)],
                                avang1.T,
                            )
                        )
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    angs1k = (
                        np.arccos(
                            np.dot(
                                orientations[(labels == kk)],
                                avang1.T,
                            )
                        )
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    angsjk = (
                        np.arccos(
                            np.dot(
                                orientations[(labels == kk)],
                                avangj.T,
                            )
                        )
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    angskj = (
                        np.arccos(
                            np.dot(
                                orientations[(labels == jj)],
                                avangk.T,
                            )
                        )
                        * 360.0
                        / (2.0 * np.pi)
                    )
                    # Debug
                    if verbose > 0 or debugH > 0:
                        sk += f"cluster combo:{ii} & {jj} & {kk} size:{N_elems},{N_elems_jj}, {N_elems_kk} \n"
                        sk += f"size check {np.abs(N_elems - (N_elems_jj + N_elems_kk))} (calced)< {(1 - pct_size_buffer) * np.max([N_elems, N_elems_jj + N_elems_kk])}\n"
                        sk += f"ang diff={np.abs(np.mean(angs1jk)-104.5):.2f}(calced)<{angdiff_cutoff:.2f},std dev:{angstd_cutoff:.2f}>{np.std(angs1jk):.2f}(calced)\n"
                        sk += f"ij {ii},{jj}, {np.abs(np.mean(angs1j)-104.5)},<,angdiff_cutoff,and std {np.std(angs1j)}<{angstd_cutoff} \n"
                        sk += f"ik {ii},{kk}, {np.abs(np.mean(angs1k)-104.5)},<,angdiff_cutoff,and std {np.std(angs1k)}<{angstd_cutoff} \n"
                        sk += f"kj {kk},{jj}, {np.abs(np.mean(angskj)-104.5)},<,angdiff_cutoff,and std {np.std(angskj)}<{angstd_cutoff} \n"
                        sk += f"jk {jj},{kk}, {np.abs(np.mean(angsjk)-104.5)},<,angdiff_cutoff,and std {np.std(angsjk)}<{angstd_cutoff} \n"
                    # check if size of clusters is about equal ii==jj+kk with angle combination
                    if (
                        np.abs(N_elems - (N_elems_jj + N_elems_kk))
                        < (1 - pct_size_buffer)
                        * np.max([N_elems, N_elems_jj + N_elems_kk])
                        and (
                            np.abs(np.mean(angs1jk) - 104.5) < angdiff_cutoff
                            and np.std(angs1jk) < angstd_cutoff
                        )
                    ) or (
                        (
                            np.abs(np.mean(angs1j) - 104.5) < angdiff_cutoff
                            and np.std(angs1j) < angstd_cutoff
                        )
                        and (
                            np.abs(np.mean(angs1k) - 104.5) < angdiff_cutoff
                            and np.std(angs1k) < angstd_cutoff
                        )
                        and (
                            np.abs(np.mean(angsjk) - 104.5) < angdiff_cutoff
                            and np.std(angsjk) < angstd_cutoff
                        )
                        and (
                            np.abs(np.mean(angskj) - 104.5) < angdiff_cutoff
                            and np.std(angskj) < angstd_cutoff
                        )
                    ):
                        weakly.append([ii, jj])
                        weakly.append([ii, kk])
                        used.append(ii)
                        used.append(jj)
                        used.append(kk)
                        break
    # check if used ones account for pct_explained percentage of orientations
    total = 0
    for i in used:
        total += len(orientations[labels == i])
    if verbose > 1:
        print("total length=", total, ">", 2 * neioc * pct_explained)
        for ws in weakly:
            print("weakly combo", ws[0], ws[1])
    if total > 2 * neioc * pct_explained:
        if verbose > 0:
            print("weakly found")
        for ws in weakly:
            weakly_conserved.append(
                __return_normalized_orientation_pair(
                    orientations, labels, ws[0], ws[1], "WCW"
                )
            )
    else:  # or if angle beteween a larger set of clusters is water angle for all of them - this means that this is circular triplet or multiplet
        used = []
        weakly = []
        # loop over OPTICS clusters
        for ii in lbls:
            this = []
            # check if in used
            if ii in used:
                continue
            # average orientation of ii
            avangi = np.mean(orientations[labels == ii], axis=0)
            # loop over clusters but not ii
            lblsni = lbls[lbls != ii]
            for jj in lblsni:
                # check if already used
                if jj in used:
                    continue
                avangj = np.mean(orientations[labels == jj], axis=0)
                # loop over clusters but not ii or jj
                lblsnij = lblsni[lblsni != jj]
                angsij = (
                    np.arccos(
                        np.dot(
                            orientations[(labels == jj)],
                            avangi.T,
                        )
                    )
                    * 360.0
                    / (2.0 * np.pi)
                )
                angsji = (
                    np.arccos(
                        np.dot(
                            orientations[(labels == ii)],
                            avangj.T,
                        )
                    )
                    * 360.0
                    / (2.0 * np.pi)
                )
                if verbose > 0 or debugH > 0:
                    sk += f"ij,{ii},{jj}, angs and std ij: {np.abs(np.mean(angsij) - 104.5)}, {np.std(angsij)}; ji: {np.abs(np.mean(angsji) - 104.5)}, {np.std(angsji)} \n"
                if (
                    np.abs(np.mean(angsij) - 104.5) < angdiff_cutoff
                    and np.std(angsij) < angstd_cutoff
                ) and (
                    np.abs(np.mean(angsji) - 104.5) < angdiff_cutoff
                    and np.std(angsji) < angstd_cutoff
                ):
                    go = True
                    for kk in this:
                        avangk = np.mean(orientations[labels == kk], axis=0)
                        angsjk = (
                            np.arccos(
                                np.dot(
                                    orientations[(labels == kk)],
                                    avangj.T,
                                )
                            )
                            * 360.0
                            / (2.0 * np.pi)
                        )
                        angskj = (
                            np.arccos(
                                np.dot(
                                    orientations[(labels == jj)],
                                    avangk.T,
                                )
                            )
                            * 360.0
                            / (2.0 * np.pi)
                        )
                        if verbose > 0 or debugH > 0:
                            sk += f"jk,{jj},{kk}, angs and std jk: {np.abs(np.mean(angsjk) - 104.5)}, {np.std(angsjk)}; kj: {np.abs(np.mean(angskj) - 104.5)}, {np.std(angskj)} \n"
                        if (
                            np.abs(np.mean(angskj) - 104.5) > angdiff_cutoff
                            and np.std(angskj) > angstd_cutoff
                        ) and (
                            np.abs(np.mean(angsjk) - 104.5) > angdiff_cutoff
                            and np.std(angsjk) > angstd_cutoff
                        ):
                            go = False
                    if go:
                        weakly.append([ii, jj])
                        used.append(ii)
                        used.append(jj)
        total = 0
        for i in used:
            total += len(orientations[labels == i])
        if verbose > 2:
            print("total length=", total, ">", 2 * neioc * pct_explained)
            for ws in weakly:
                print("weakly combo", ws[0], ws[1])
        if total > 2 * neioc * pct_explained:
            if verbose > 0:
                print("weakly found")
            for ws in weakly:
                weakly_conserved.append(
                    __return_normalized_orientation_pair(
                        orientations, labels, ws[0], ws[1], "WCW"
                    )
                )

    # Debug
    if verbose > 0 or debugH > 0:
        ss = (
            f"weakly Conserved - OPTICS analysis;xi={xi}\n"
            f"Number of hydrogen clusters : {len(np.unique(labels))};\n number of elements : {counts}; range needed for best cluster:(depends on numbpct) {neioc*(2-pct_size_buffer):.2f},{neioc*lower_bound_pct_buffer:.2f}\n"
        )
    # Debug Printing
    if verbose == 2 or (verbose == 1 and len(weakly_conserved) > 0):
        print(ss + sk)
    # Debug plots
    __hydrogen_orient_plots(
        labels,
        orientations,
        cc,
        ss,
        f"Reachability of optics plot\n minsamples={int(neioc*min_samp_data_size_pct)}; xi={xi}",
        len(weakly_conserved) > 0,
        debugH,
        plotreach,
    )
    return weakly_conserved


def __hydrogen_orient_plots(
    labels, orientations, cc, ss, rtit, conserved, debugH, plotreach: bool
) -> None:
    """Collection of plots for hydrogen orientation.

    For debuging purposes. Not ment for direct usage.
    """
    if debugH == 2 or (debugH == 1 and conserved):
        if plotreach:
            fig = __plot3Dorients(111, labels, orientations, ss)
        else:
            fig = __plot3Dorients(121, labels, orientations, ss)
        if plotreach:
            __plotreachability(122, orientations, cc, fig=fig, tit=rtit)


def __plot3Dorients(subplot, labels, orientations, tip) -> Figure:
    """Function for plotting 3D orientations.

    For debuging only.

    """
    try:
        import matplotlib.pyplot as plt
    except ModuleNotFoundError:
        raise Exception("install matplotlib")

    fig: Figure = plt.figure()
    if type(labels) == int:
        return fig
    ax: Axes = fig.add_subplot(subplot, projection="3d")
    ax.set_title(tip)
    for j in np.unique(labels):
        jaba = orientations[labels == j]
        ax.scatter(
            jaba[:, 0],
            jaba[:, 1],
            jaba[:, 2],
            label=f"{j} ({len(labels[labels==j])})",
        )
        if j > -1:
            ax.quiver(
                0,
                0,
                0,
                np.mean(jaba[:, 0]),
                np.mean(jaba[:, 1]),
                np.mean(jaba[:, 2]),
                color="gray",
                arrow_length_ratio=0.0,
                linewidths=5,
            )
    ax.scatter(0, 0, 0, c="crimson", s=1000)
    ax.legend()
    ax.grid(False)
    ax.axis("off")
    ax.set_aspect("equal")
    ax.autoscale(tight=True)
    ax.dist = 6
    return fig


def __plotreachability(
    subplot, orientations, cc, fig: Figure | None = None, tit=None
) -> Figure:
    """Function for plotting reachability.

    For debuging purposes only.

    """
    try:
        import matplotlib.pyplot as plt
    except ModuleNotFoundError:
        raise Exception("install matplotlib")

    if fig is None:
        fig: Figure = plt.figure()
    if type(cc) != OPTICS:
        return fig
    lblls = cc.labels_[cc.ordering_]
    labels = cc.labels_
    reachability = cc.reachability_[cc.ordering_]
    ax2 = fig.add_subplot(subplot)
    fig.gca().set_prop_cycle(None)
    space = np.arange(len(orientations))
    ax2.plot(space, reachability)
    if tit is not None:
        ax2.set_title(tit)
    for clst in np.unique(lblls):
        if clst == -1:
            ax2.plot(
                space[lblls == clst],
                reachability[lblls == clst],
                label=f"{clst} ({len(space[lblls==clst])}), avg reach={np.mean(np.ma.masked_invalid(cc.reachability_[labels==clst]))}",
                color="blue",
            )
        else:
            ax2.plot(
                space[lblls == clst],
                reachability[lblls == clst],
                label=f"{clst} ({len(space[lblls==clst])}), avg reach={np.mean(np.ma.masked_invalid(cc.reachability_[labels==clst]))}",
            )
    ax2.legend()
    return fig


def __return_normalized_orientation_pair(
    orientations: np.ndarray,
    labels: np.ndarray,
    i: int,
    j: int,
    typel: str,
) -> list[np.ndarray | str]:
    """Helper function for normalizing orientations

    Not ment for general usage.

    """
    v1 = np.mean(orientations[labels == i], axis=0)
    v2 = np.mean(orientations[labels == j], axis=0)
    return [
        v1 / np.linalg.norm(v1) * np.linalg.norm(orientations[0]),
        v2 / np.linalg.norm(v2) * np.linalg.norm(orientations[0]),
        typel,
    ]
